"""Some implemented mixins."""

import typing as t

import numpy as np
import numpy.typing as npt

from taurex.chemistry import Gas
from taurex.core import FittingType, fitparam

from . import ChemistryMixin, TemperatureMixin


class MakeFreeMixin(ChemistryMixin):
    """Provides a :func:`addGas` method to any chemistry

    This class will either inject or force a molecule
    to become a :class:`~taurex.data.profiles.chemistry.gas.gas.Gas`
    object. Allowing them to be freely changed or retrieved.

    For example lets enhance ACE:

    >>> from acepython.taurex import ACEChemistry
    >>> from taurex.mixin import enhance_class, MakeFreeMixin
    >>> old_ace = ACEChemistry()
    >>> new_ace = enhance_class(ACEChemistry, MakeFreeMixin)

    ``new_ace`` behaves the same as ``old_ace``:

    >>> new_ace.ace_metallicity
    1.0

    And we see the same molecules and fitting parameters exist:

    >>> old_ace.gases == new_ace.gases
    True
    >>> new_ace.gases
    ['CH3COOOH', 'C4H9O', ... 'HNC', 'HON', 'NCN']
    >>> new_ace.fitting_parameters().keys()
    dict_keys(['ace_metallicity', 'metallicity', 'ace_co', 'C_O_ratio'])

    ``new_ace`` is embued with the

    >>> from taurex.chemistry import ConstantGas
    >>> new_ace.addGas(ConstantGas('TiO',1e-8)).addGas(ConstantGas('VO',1e-8))
    >>> new_ace.gases == old_ace.gases
    False
    >>> new_ace.gases
    ['CH3COOOH', 'C4H9O', ... 'HNC', 'HON', 'NCN', 'TiO', 'VO']

    And indeed see that they are included. We can also retrieve them:

    >>> new_ace.fitting_parameters().keys()
    dict_keys(['TiO', 'VO', 'ace_metallicity', 'metallicity', 'ace_co', 'C_O_ratio'])

    Finally we can force an existing molecule like CH4 into becoming a Gas:

    >>> new_ace.addGas(ConstantGas('CH4',1e-5))

    And see that it is now a retrieval parameter as well.

    >>> new_ace.fitting_parameters().keys()
    dict_keys(['TiO', 'VO', 'CH4', 'ace_metallicity',
                'metallicity', 'ace_co', 'C_O_ratio'])



    """

    def __init_mixin__(self) -> None:
        """Initialise free chemistry mixin."""
        from taurex.chemistry import TaurexChemistry

        if isinstance(self, TaurexChemistry):
            raise ValueError("Class is already free-type")
        self._mixin_new_gas_list: t.List[Gas] = []
        self.active_exist: t.List[Gas] = []
        self.inactive_exist: t.List[Gas] = []
        self.active_nonexist: t.List[Gas] = []
        self.inactive_nonexist: t.List[Gas] = []
        self.show_old_gases = True

    @property
    def gases(self) -> t.List[str]:
        """Gases in the atmosphere + injected."""
        if not hasattr(self, "show_old_gases"):
            self.show_old_gases = True
        if self.show_old_gases:
            return super().gases
        else:
            return (
                super().gases
                + [g.molecule for g in self.active_nonexist]
                + [g.molecule for g in self.inactive_nonexist]
            )

    def addGas(self, gas: Gas) -> "MakeFreeMixin":  # noqa: N802
        """Adds a gas in the atmosphere.

        Parameters
        ----------
        gas : :class:`~taurex.data.profiles.chemistry.gas.gas.Gas`
            Gas to add into the atmosphere. Only takes effect
            on next initialization call.

        """

        if gas.molecule in [x.molecule for x in self._mixin_new_gas_list]:
            self.error("Gas already exists %s", gas.molecule)
            raise ValueError("Gas already exists")

        self._mixin_new_gas_list.append(gas)
        self.determine_new_mix_mask()
        self.norm_factor = 1.0

        return self

    def determine_new_mix_mask(self) -> None:
        """Determines which gases are new and which are old."""
        current_gases = self._mixin_new_gas_list
        self.active_exist = []
        self.inactive_exist = []
        self.active_nonexist = []
        self.inactive_nonexist = []
        self.show_old_gases = True
        current_active = super().activeGases
        current_inactive = super().inactiveGases
        self.show_old_gases = False
        for g in current_gases:
            if g.molecule in current_active:
                self.active_exist.append((g, current_active.index(g.molecule)))
            elif g.molecule in current_inactive:
                self.inactive_exist.append((g, current_inactive.index(g.molecule)))
            elif g.molecule in super().availableActive:
                self.active_nonexist.append(g)
            else:
                self.inactive_nonexist.append(g)

    @property
    def activeGases(self) -> t.List[str]:  # noqa: N802
        """Active gases in the atmosphere + injected."""
        return list(super().activeGases) + [g.molecule for g in self.active_nonexist]

    @property
    def inactiveGases(self) -> t.List[str]:  # noqa: N802
        """Inactive gases in the atmosphere + injected."""
        return list(super().inactiveGases) + [
            g.molecule for g in self.inactive_nonexist
        ]

    @property
    def activeGasMixProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Active gas layer by layer mix profile.

        Returns
        -------
        active_mix_profile : :obj:`array`

        """
        mix_profile = super().activeGasMixProfile

        if mix_profile is not None:
            for g, idx in self.active_exist:
                mix_profile[idx] = g.mixProfile

        if len(self.active_nonexist) > 0:
            nonexist_profile = np.array([g.mixProfile for g in self.active_nonexist])
            if mix_profile is None:
                return nonexist_profile
            else:
                return (
                    np.concatenate((mix_profile, nonexist_profile)) / self.norm_factor
                )
        else:
            if mix_profile is None:
                return None
            return mix_profile / self.norm_factor

    @property
    def inactiveGasMixProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Inactive gas layer by layer mix profile

        Returns
        -------
        inactive_mix_profile : :obj:`array`

        """
        mix_profile = super().inactiveGasMixProfile
        for g, idx in self.inactive_exist:
            mix_profile[idx] = g.mixProfile

        if len(self.inactive_nonexist) > 0:
            nonexist_profile = np.array([g.mixProfile for g in self.inactive_nonexist])
            return np.concatenate((mix_profile, nonexist_profile)) / self.norm_factor
        else:
            return mix_profile / self.norm_factor

    def initialize_chemistry(
        self,
        nlayers: int,
        temperature_profile: npt.NDArray[np.float64],
        pressure_profile: npt.NDArray[np.float64],
        altitude_profile: t.Optional[npt.NDArray[np.float64]] = None,
    ) -> None:
        """Initialize the chemistry.

        Parameters
        -----------
        nlayers : int
            number of layers
        temperature_profile : np.ndarray
            temperature profile
        pressure_profile : np.ndarray
            pressure profile
        altitude_profile : np.ndarray , optional
            altitude profile, deprecated

        """
        self._run = False
        super().initialize_chemistry(
            nlayers, temperature_profile, pressure_profile, altitude_profile
        )

        for g in self._mixin_new_gas_list:
            g.initialize_profile(
                nlayers, temperature_profile, pressure_profile, altitude_profile
            )
        self._run = True
        self.norm_factor = 1.0

        active_norm = 0.0
        inactive_norm = 0.0
        active_mix = self.activeGasMixProfile
        inactive_mix = self.inactiveGasMixProfile

        if active_mix is not None:
            active_norm = np.sum(self.activeGasMixProfile, axis=0)
        if inactive_mix is not None:
            inactive_norm = np.sum(self.inactiveGasMixProfile, axis=0)

        self.norm_factor = active_norm + inactive_norm
        self.compute_mu_profile(nlayers)

    def compute_mu_profile(self, nlayers: int):
        """Computes molecular weight of atmosphere for each layer.


        Parameters
        ----------
        nlayers: int
            Number of layers, deprecated
        """
        from taurex.util import get_molecular_weight

        if not self._run:
            return
        self.show_old_gases = True
        super().compute_mu_profile(nlayers)
        self.show_old_gases = False
        self._mu_profile = super().muProfile

        for idx, g in enumerate(reversed(self.active_nonexist)):
            self._mu_profile += (
                get_molecular_weight(g.molecule) * self.activeGasMixProfile[-idx - 1]
            )

        for idx, g in enumerate(reversed(self.inactive_nonexist)):
            self._mu_profile += (
                get_molecular_weight(g.molecule) * self.inactiveGasMixProfile[-idx - 1]
            )

    def fitting_parameters(self) -> t.Dict[str, FittingType]:
        """Adds fitting parameters for injected gases.

        Overrides the fitting parameters to return
        one with all the gas profile parameters as well

        Returns
        -------

        fit_param : :obj:`dict`

        """
        full_dict = {}
        for gas in self._mixin_new_gas_list:
            full_dict.update(gas.fitting_parameters())

        full_dict.update(self._param_dict)

        return full_dict

    @property
    def muProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Molecular weight profile.

        Returns
        -------
        mu_profile : :obj:`array`

        """
        return self._mu_profile

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Input keywords for mixin."""
        return ("makefree",)


class TempScaler(TemperatureMixin):
    """Scales the temperature profile by a factor.

    Mostly a proof of concept class for mixins. Not particularly
    useful in practice.

    """

    def __init_mixin__(self, scale_factor: t.Optional[float] = 1.0) -> None:
        """Initialise the temperature scaler."""
        self._scale_factor = scale_factor

    @fitparam(param_name="T_scale")
    def scaleFactor(self) -> float:  # noqa: N802
        """Scale factor for temperature profile."""
        return self._scale_factor

    @scaleFactor.setter
    def scaleFactor(self, value: float) -> None:  # noqa: N802
        """Sets the scale factor."""
        self._scale_factor = value

    @property
    def profile(self) -> npt.NDArray[np.float64]:
        """Scaled temperature profile."""
        return super().profile * self._scale_factor

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Input keywords for mixin."""
        return ("tempscalar",)



"""Handles how priors are defined for retrievals."""
import enum
import math
import typing as t

import scipy.stats as stats

from taurex.log import Logger


class PriorMode(enum.Enum):
    """Defines the type of prior space."""

    LINEAR = 0
    LOG = 1


class Prior(Logger):
    """Defines a prior function"""

    def __init__(self) -> None:
        """Initialise prior"""
        super().__init__(self.__class__.__name__)
        self._prior_mode = PriorMode.LINEAR

    @property
    def priorMode(self) -> PriorMode:  # noqa: N802
        """Defined prior mode."""
        return self._prior_mode

    def sample(self, x: float) -> float:
        """Sample the prior.

        Parameters
        ----------
        x : float
            A random number between 0 and 1.

        Raises
        ------
        NotImplementedError
            [description]

        """
        raise NotImplementedError

    def prior(self, value: float) -> float:
        """Convert a value from prior space to linear space."""
        if self._prior_mode is PriorMode.LINEAR:
            return value
        else:
            return 10**value

    def params(self) -> str:
        """Return the parameters of the prior in string form.

        Raises:
            NotImplementedError: [description]

        """
        raise NotImplementedError

    def boundaries(self) -> t.Tuple[float, float]:
        """Return the (rough) boundaries of the prior.

        Raises:
            NotImplementedError: [description]

        """
        raise NotImplementedError


class Uniform(Prior):
    """Defines a uniform prior."""

    def __init__(
        self,
        bounds: t.Optional[t.Tuple[float, float]] = (
            0.0,
            1.0,
        ),
    ):
        """Initialise uniform prior.

        Parameters
        ----------
        bounds : tuple, optional
            Bounds of the prior, by default (0.0, 1.0)

        Raises
        ------
        ValueError
            If no bounds are defined.

        """
        super().__init__()
        if bounds is None:
            self.error("No bounds defined")
            raise ValueError("No bounds defined")

        self.set_bounds(bounds)

    def set_bounds(self, bounds: t.Tuple[float, float]):
        """Set the bounds of the prior.

        Parameters
        ----------
        bounds : t.Tuple[float, float]
            Bounds of the prior.

        """
        self._low_bounds = min(*bounds)
        self.debug("Lower bounds: %s", self._low_bounds)
        self._up_bounds = max(*bounds)
        self._scale = self._up_bounds - self._low_bounds
        self.debug("Scale: %s", self._scale)

    def sample(self, x: float) -> float:
        """Sample the prior.

        Parameters
        ----------
        x : float
            A random number between 0 and 1.

        """
        return stats.uniform.ppf(x, loc=self._low_bounds, scale=self._scale)

    def params(self):
        """Return the parameters of the prior in string form.

        Returns
        -------
        str
            String representation of the prior.

        """
        return f"Bounds = [{self._low_bounds},{self._up_bounds}]"

    def boundaries(self) -> t.Tuple[float, float]:
        """Return the boundaries of the prior."""
        return self._low_bounds, self._up_bounds


class LogUniform(Uniform):
    """Defines a log-uniform prior."""

    def __init__(
        self,
        bounds: t.Optional[t.Tuple[float, float]] = (
            0.0,
            1.0,
        ),
        lin_bounds=None,
    ):
        """Initialise log-uniform prior.

        Parameters
        ----------
        bounds : list, optional
            Bounds of the prior in logspace, by default [0.0, 1.0]

        lin_bounds : list, optional
            Bounds of the prior in linear space, by default None

        """
        if lin_bounds is not None:
            bounds = [math.log10(x) for x in lin_bounds]
        super().__init__(bounds=bounds)
        self._prior_mode = PriorMode.LOG


class Gaussian(Prior):
    """Defines a gaussian prior."""

    def __init__(self, mean: t.Optional[float] = 0.5, std: t.Optional[float] = 0.25):
        """Initialise gaussian prior.

        Parameters
        ----------
        mean : float, optional
            Mean of the gaussian, by default 0.5
        std : float, optional
            Standard deviation of the gaussian, by default 0.25

        """
        super().__init__()

        self._loc = mean
        self._scale = std

    def sample(self, x: float) -> float:
        """Sample the prior.

        Parameters
        ----------
        x : float
            A random number between 0 and 1.

        """
        return stats.norm.ppf(x, loc=self._loc, scale=self._scale)

    def params(self) -> str:
        """Return the parameters of the prior in string form."""
        return f"Mean = {self._loc} Stdev = {self._scale}"

    def boundaries(self) -> t.Tuple[float, float]:
        """Return the boundaries of the prior."""
        return self.sample(0.1), self.sample(0.9)


class LogGaussian(Gaussian):
    """Defines a log-gaussian prior."""

    def __init__(
        self,
        mean: t.Optional[float] = 0.5,
        std: t.Optional[float] = 0.25,
        lin_mean: t.Optional[float] = None,
        lin_std: t.Optional[float] = None,
    ):
        """Initialise log-gaussian prior.

        Parameters
        ----------
        mean : float, optional
            Mean of the gaussian, by default 0.5
        std : float, optional
            Standard deviation of the gaussian, by default 0.25
        lin_mean : float, optional
            Mean of the gaussian in linear space, by default None
        lin_std : float, optional
            Standard deviation of the gaussian in linear space, by default None

        """
        if lin_mean is not None:
            mean = math.log10(lin_mean)
        if lin_std is not None:
            std = math.log10(lin_std)
        super().__init__(mean=mean, std=std)
        self._prior_mode = PriorMode.LOG



"""Spectra from Iraclis pickle data."""
import pickle  # noqa: S403
import typing as t

import numpy as np

from taurex.types import PathLike

from .array import ArraySpectrum


class IraclisSpectrum(ArraySpectrum):
    """Loads an observation from Iraclis pickle data."""

    def __init__(self, filename: t.Optional[PathLike] = None):
        """Initialize.

        Parameters
        ----------
        filename:
            Filename of Iraclis pickle data, by default None


        """
        self._filename = filename
        try:
            with open(filename, "rb") as f:
                database = pickle.load(f)  # noqa: S301
        except UnicodeDecodeError:
            with open(filename, "rb") as f:
                database = pickle.load(f, encoding="latin1")  # noqa: S301

        wl = database["spectrum"]["wavelength"]
        td = database["spectrum"]["depth"]
        err = database["spectrum"]["error"]
        width = database["spectrum"]["width"]

        final_array = np.vstack((wl, td, err, width)).T

        super().__init__(final_array)

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        return ("iraclis",)



"""Opacity module using RADIS code"""
import typing as t
from functools import lru_cache

import astropy.units as u
import numpy as np
import numpy.typing as npt
from astropy.constants import k_B

from .opacity import Opacity


class RadisHITRANOpacity(Opacity):
    """Radis opacity using HITRAN database.

    This module uses the RADIS code to compute opacity
    using the HITRAN database. This module is not recommended
    for general use as it is very slow and is deprioritized
    in the opacity module list.

    """

    @classmethod
    def priority(cls) -> int:
        """Priority of opacity module."""
        return 1000000

    @classmethod
    def discover(cls) -> t.List[t.Tuple[str, t.Tuple[str, float, float, int]]]:
        from taurex.cache import GlobalCache

        if GlobalCache()["enable_radis"] is not True:
            return []

        trans = {
            "1": "H2O",
            "2": "CO2",
            "3": "O3",
            "4": "N2O",
            "5": "CO",
            "6": "CH4",
            "7": "O2",
            "9": "SO2",
            "10": "NO2",
            "11": "NH3",
            "12": "HNO3",
            "13": "OH",
            "14": "HF",
            "15": "HCl",
            "16": "HBr",
            "17": "HI",
            "18": "ClO",
            "19": "OCS",
            "20": "H2CO",
            "21": "HOCl",
            "23": "HCN",
            "24": "CH3Cl",
            "25": "H2O2",
            "26": "C2H2",
            "27": "C2H6",
            "28": "PH3",
            "29": "COF2",
            "30": "SF6",
            "31": "H2S",
            "32": "HCOOH",
            "33": "HO2",
            "34": "O",
            "35": "ClONO2",
            "36": "NO+",
            "37": "HOBr",
            "38": "C2H4",
            "40": "CH3Br",
            "41": "CH3CN",
            "42": "CF4",
            "43": "C4H2",
            "44": "HC3N",
            "46": "CS",
            "47": "SO3",
        }

        mol_list = trans.values()
        wn_start, wn_end, wn_points = 600, 30000, 10000
        grid = GlobalCache()["radis_grid"]
        if grid is not None:
            wn_start, wn_end, wn_points = grid

        return [(m, [m, wn_start, wn_end, wn_points]) for m in mol_list]

    def __init__(
        self,
        molecule_name: str,
        wn_start: t.Optional[float] = 600,
        wn_end: t.Optional[float] = 30000,
        wn_points: t.Optional[int] = 10000,
    ):
        """Initialize RadisHITRANOpacity.

        Parameters
        ----------
        molecule_name:
            Molecule name.

        wn_start:
            Wavenumber start, by default 600 cm-1.

        wn_end:
            Wavenumber end, by default 30000 cm-1.

        wn_points:
            Wavenumber points, by default 10000.

        """
        super().__init__(self.__class__.__name__)
        import radis

        step = (wn_end - wn_start) / wn_points

        self.info("RADIS Grid set to %s %s %s", wn_start, wn_end, step)

        self.rad_xsec = radis.SpectrumFactory(
            wavenum_min=wn_start,
            wavenum_max=wn_end,
            isotope="1",  # 'all',
            # depends on HAPI benchmark.
            wstep=step,
            verbose=0,
            cutoff=1e-27,
            mole_fraction=1.0,
            # Corresponds to WavenumberWingHW/HWHM=50 in HAPI
            broadening_max_width=10.0,
            molecule=molecule_name,
        )

        self.rad_xsec.fetch_databank("astroquery", load_energies=False)

        self._molecule_name = molecule_name

        s = self.rad_xsec.eq_spectrum(Tgas=296, pressure=1)
        wn, absnce = s.get("abscoeff")

        self.wn = 1e7 / wn

    @property
    def moleculeName(self) -> str:  # noqa: N802
        """Molecule name."""
        return self._molecule_name

    @property
    def wavenumberGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Wavenumber grid."""
        return self.wn

    @property
    def temperatureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Temperature grid."""
        raise NotImplementedError

    @property
    def pressureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Pressure grid."""
        raise NotImplementedError

    @lru_cache(maxsize=500)  # noqa: B019
    def compute_opacity(
        self, temperature: float, pressure: float
    ) -> npt.NDArray[np.float64]:
        """Compute opacity.

        Parameters
        ----------
        temperature:
            Temperature in K.

        pressure:
            Pressure in bar.

        Returns
        -------
        opacity:
            Opacity array at K, Pa.

        """
        pressure_pascal = pressure * u.Pascal
        pressure_bar = pressure_pascal.to(u.bar)
        temperature_k = temperature * u.K

        density = ((pressure_bar) / (k_B * temperature_k)).value * 1000

        s = self.rad_xsec.eq_spectrum(Tgas=temperature, pressure=pressure_bar.value)
        wn, absnce = s.get("abscoeff")

        return absnce / density

    BIBTEX_ENTRIES = [
        r"""
        @ARTICLE{2019JQSRT.222...12P,
            author = {{Pannier}, Erwan and {Laux}, Christophe O.},
                title = "{RADIS: A nonequilibrium
                line-by-line radiative code for CO$_{2}$ and
                HITRAN-like database species}",
            journal = {Journal of Quantitative Spectroscopy & Radiative Transfer},
            keywords = {Line-by-line code, Nonequilibrium,
            Optical emission spectroscopy, Absorption spectroscopy},
                year = 2019,
                month = jan,
            volume = {222},
                pages = {12-25},
                doi = {10.1016/j.jqsrt.2018.09.027},
            adsurl = {https://ui.adsabs.harvard.edu/abs/2019JQSRT.222...12P},
            adsnote = {Provided by the SAO/NASA Astrophysics Data System}
        }


        """,
    ]



"""Base and Concrete pressure profiles."""

import math
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.data.citation import Citable
from taurex.data.fittable import Fittable, fitparam
from taurex.log import Logger
from taurex.output import OutputGroup
from taurex.output.writeable import Writeable


class PressureProfile(Fittable, Logger, Writeable, Citable):
    """Base pressure class.

    *Abstract Class*

    Simple. Defines the layering
    of the atmosphere. Only requires
    implementation of:

    - :func:`compute_pressure_profile`
    - :func:`profile`

    """

    def __init__(self, name: str, nlayers: int) -> None:
        """Initialize pressure profile.



        Parameters
        ----------

        name: str
            Name used in logging

        nlayers: int
            Number of layers in atmosphere

        """
        Fittable.__init__(self)
        Logger.__init__(self, name)
        self.pressure_profile_levels: npt.NDArray[np.float64] = None
        if nlayers <= 0:
            self.error("Number of layers: [%s] should be greater than 0", nlayers)
            raise ValueError("Number of layers should be at least 1")

        self._nlayers = int(nlayers)

    @property
    def nLayers(self) -> int:  # noqa: N802
        """Number of central layers.

        Returns
        -------
        int
        """

        return self._nlayers

    @property
    def nLevels(self) -> int:  # noqa: N802
        """Number of levels (interface between layers)."""
        return self.nLayers + 1

    def compute_pressure_profile(self) -> None:
        """Compute pressure profile in Pa.


        **Requires implementation**

        Compute pressure profile and
        generate pressure array in Pa

        Returns
        -------
        pressure_profile: :obj:`array`
            Pressure profile array in Pa

        """
        raise NotImplementedError

    @property
    def profile(self) -> npt.NDArray[np.float64]:
        """Pressure at each atmospheric layer (Pascal)

        Returns
        -------
        pressure_profile : :obj:`array`
            Pressure profile array in Pa
        """
        raise NotImplementedError

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write pressure profile to output."""
        pressure = output.create_group("Pressure")
        pressure.write_string("pressure_type", self.__class__.__name__)
        pressure.write_scalar("nlayers", self._nlayers)
        pressure.write_array("profile", self.profile)
        return pressure


class SimplePressureProfile(PressureProfile):
    """A basic pressure profile."""

    WARN = True

    def __init__(
        self,
        nlayers: t.Optional[int] = 100,
        atm_min_pressure: t.Optional[float] = 1e-4,
        atm_max_pressure: t.Optional[float] = 1e6,
    ):
        """Initialize pressure profile.

        Parameters
        ----------
        nlayers : int
            Number of layers in atmosphere

        atm_min_pressure : float
            minimum pressure in Pascal (top of atmosphere)

        atm_max_pressure : float
            maximum pressure in Pascal (surface of planet)

        """
        from warnings import warn

        super().__init__("pressure_profile", nlayers)
        self.pressure_profile = None
        if self.WARN:
            warn(
                "SimplePressureProfile is deprecated. "
                "Use LogPressureProfile instead",
                DeprecationWarning,
            )
        if atm_max_pressure <= atm_min_pressure:
            self.error(
                "Max pressure %1.2e should be greater " "than min pressure %1.2e",
                atm_max_pressure,
                atm_min_pressure,
            )
            raise ValueError("Max pressure is less than minimum pressure")

        self._atm_min_pressure = atm_min_pressure
        self._atm_max_pressure = atm_max_pressure

    def compute_pressure_profile(self) -> None:
        """Set up the pressure profile for the atmosphere model."""

        # set pressure profile of layer boundaries
        # press_exp = np.linspace(np.log(self._atm_min_pressure),
        #                       np.log(self._atm_max_pressure),
        #                       self.nLevels)
        # self.pressure_profile_levels = np.exp(press_exp)[::-1]
        self.pressure_profile_levels = np.logspace(
            math.log10(self._atm_min_pressure),
            math.log10(self._atm_max_pressure),
            self.nLevels,
        )[::-1]
        # get mid point pressure between levels (i.e. get layer pressure)
        # computing geometric
        # average between pressure at n and n+1 level
        self.pressure_profile = self.pressure_profile_levels[:-1] * np.sqrt(
            self.pressure_profile_levels[1:] / self.pressure_profile_levels[:-1]
        )

    @fitparam(
        param_name="atm_min_pressure",
        param_latex=r"$P_\mathrm{min}$",
        default_mode="log",
        default_fit=False,
        default_bounds=[0.1, 1.0],
    )
    def minAtmospherePressure(self) -> float:  # noqa: N802
        """Minimum pressure of atmosphere (top layer) in Pascal"""
        return self._atm_min_pressure

    @minAtmospherePressure.setter
    def minAtmospherePressure(self, value: float) -> None:  # noqa: N802
        self._atm_min_pressure = value

    @fitparam(
        param_name="atm_max_pressure",
        param_latex=r"$P_\mathrm{max}$",
        default_mode="log",
        default_fit=False,
        default_bounds=[0.1, 1.0],
    )
    def maxAtmospherePressure(self) -> float:  # noqa: N802
        """Maximum pressure of atmosphere (surface) in Pascal."""
        return self._atm_max_pressure

    @maxAtmospherePressure.setter
    def maxAtmospherePressure(self, value: float):  # noqa: N802
        """Set the maximum pressure of the atmosphere (surface) in Pascal"""
        self._atm_max_pressure = value

    @property
    def profile(self) -> npt.NDArray[np.float64]:
        """Pressure at each atmospheric layer (Pascal)"""
        return self.pressure_profile

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write pressure profile to output."""
        pressure = super().write(output)

        pressure.write_scalar("atm_max_pressure", self._atm_max_pressure)
        pressure.write_scalar("atm_min_pressure", self._atm_min_pressure)

        return pressure

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        return (
            "simple",
            "hydrostatic",
            "logpressure",
        )


class LogPressureProfile(SimplePressureProfile):
    """A pressure profile built from a logspace."""

    WARN = False
    pass



"""Common functions that are used and are quite helpful"""

from .util import (bindown, calculate_weight, class_for_name,
                   class_from_keyword, clip_native_to_wngrid,
                   compute_bin_edges, compute_dz, conversion_factor,
                   create_grid_res, decode_string_array, ensure_string_utf8,
                   find_closest_pair, get_molecular_weight, has_duplicates,
                   loadtxt2d, mass, merge_elements, molecule_texlabel,
                   movingaverage, quantile_corner, random_int_iter,
                   read_error_into_dict, read_error_line, read_table,
                   recursively_save_dict_contents_to_output,
                   sanitize_molecule_string, split_molecule_elements,
                   store_thing, tokenize_molecule, weighted_avg_and_std,
                   wnwidth_to_wlwidth)

__all__ = [
    "mass",
    "calculate_weight",
    "split_molecule_elements",
    "tokenize_molecule",
    "merge_elements",
    "split_molecule_elements",
    "sanitize_molecule_string",
    "get_molecular_weight",
    "molecule_texlabel",
    "bindown",
    "movingaverage",
    "quantile_corner",
    "loadtxt2d",
    "read_error_line",
    "read_error_into_dict",
    "read_table",
    "decode_string_array",
    "recursively_save_dict_contents_to_output",
    "store_thing",
    "weighted_avg_and_std",
    "random_int_iter",
    "compute_bin_edges",
    "clip_native_to_wngrid",
    "wnwidth_to_wlwidth",
    "class_from_keyword",
    "class_for_name",
    "create_grid_res",
    "conversion_factor",
    "compute_dz",
    "has_duplicates",
    "find_closest_pair",
    "ensure_string_utf8",
]



"""Module for all things planet."""
import math
import typing as t
from warnings import warn

import numpy as np
import numpy.typing as npt

from taurex.constants import AU, MJUP, RJUP, G
from taurex.log import Logger
from taurex.output import OutputGroup
from taurex.output.writeable import Writeable
from taurex.util import conversion_factor

from .citation import Citable
from .fittable import Fittable, derivedparam, fitparam


class BasePlanet(Fittable, Logger, Writeable, Citable):
    """Base Planet class.

    Holds information on a planet and its properties and
    derived properties

    """

    def __init__(
        self,
        planet_mass: t.Optional[float] = 1.0,
        planet_radius: t.Optional[float] = 1.0,
        planet_sma: t.Optional[float] = None,
        planet_distance: t.Optional[float] = 1.0,
        impact_param: t.Optional[float] = 0.5,
        orbital_period: t.Optional[float] = 2.0,
        albedo: t.Optional[float] = 0.3,
        transit_time: t.Optional[float] = 3000.0,
    ) -> None:
        """Initialise a planet.

        Parameters
        -----------

        planet_mass: float, optional
            mass in terms of Jupiter mass of the planet
        planet_radius: float, optional
            radius in terms of Jupiter radii of the planet
        planet_sma: float, optional
            Semi-major axis in AU
        impact_param: float, optional
            Impact parameter
        orbital_period: float, optional
            Orbital period in days
        albedo: float, optional
            Planetary albedo
        transit_time: float, optional
            Transit time in seconds

        """
        Logger.__init__(self, "Planet")
        Fittable.__init__(self)
        self.set_planet_mass(planet_mass, "Mjup")
        self.set_planet_radius(planet_radius, "Rjup")
        self.set_planet_semimajoraxis(planet_sma or planet_distance)
        self._impact = impact_param
        self._orbit_period = orbital_period
        self._albedo = albedo
        self._transit_time = transit_time

    def set_planet_radius(self, value: float, unit: t.Optional[str] = "Rjup") -> None:
        """Set planet radius."""
        factor = conversion_factor(unit, "m")
        self._radius = value * factor

    def set_planet_mass(self, value: float, unit="Mjup") -> None:
        """Set planet mass."""
        factor = conversion_factor(unit, "kg")
        self._mass = value * factor

    def set_planet_semimajoraxis(self, value: float, unit="AU") -> None:
        """Set planet semi major axis."""
        factor = conversion_factor(unit, "m")
        self._distance = value * factor

    def get_planet_radius(self, unit: t.Optional[str] = "Rjup") -> float:
        """Get planet radius in specified unit (default is Rjup)."""
        factor = conversion_factor("m", unit)
        return self._radius * factor

    def get_planet_mass(self, unit: t.Optional[str] = "Mjup") -> float:
        """Get planet mass in specified unit (default is Mjup)."""
        factor = conversion_factor("kg", unit)
        return self._mass * factor

    def get_planet_semimajoraxis(self, unit: t.Optional[str] = "AU") -> float:
        """Get planet semi major axis in specified unit (default is AU)."""
        factor = conversion_factor("m", unit)
        return self._distance * factor

    @fitparam(
        param_name="planet_mass",
        param_latex="$M_p$",
        default_fit=False,
        default_bounds=[0.5, 1.5],
    )
    def mass(self) -> float:
        """Planet mass in Jupiter mass."""
        return self.get_planet_mass(unit="Mjup")

    @mass.setter
    def mass(self, value: float) -> None:
        """Set planet mass in Jupiter mass."""
        self.set_planet_mass(value, unit="Mjup")

    @fitparam(
        param_name="planet_radius",
        param_latex="$R_p$",
        default_fit=False,
        default_bounds=[0.9, 1.1],
    )
    def radius(self) -> float:
        """Planet radius in Jupiter radii."""
        return self.get_planet_radius(unit="Rjup")

    @radius.setter
    def radius(self, value: float) -> None:
        """Set planet radius in Jupiter radii."""
        self.set_planet_radius(value, unit="Rjup")

    @property
    def fullRadius(self) -> float:  # noqa: N802
        """Planet radius in metres

        Deprecated, use :func:`get_planet_radius` instead

        """
        warn(
            "fullRadius is deprecated, use get_planet_radius(unit='m') instead",
            DeprecationWarning,
        )
        return self.get_planet_radius(unit="m")

    @property
    def fullMass(self) -> float:  # noqa: N802
        """Planet mass in kg.

        Deprecated, use :func:`get_planet_mass` instead

        """
        warn(
            "fullMass is deprecated, use get_planet_mass(unit='kg') instead",
            DeprecationWarning,
        )
        return self._mass

    @property
    def impactParameter(self) -> float:  # noqa: N802
        """Planet impact parameter."""
        return self._impact

    @property
    def orbitalPeriod(self) -> float:  # noqa: N802
        """Planet orbital period in days."""
        return self._orbit_period

    @property
    def albedo(self) -> float:  # noqa: N802
        """Planet albedo."""
        return self._albedo

    @property
    def transitTime(self) -> float:  # noqa: N802
        """Planet transit time in seconds."""
        return self._transit_time

    @fitparam(
        param_name="planet_distance",
        param_latex="$D_{planet}$",
        default_fit=False,
        default_bounds=[1, 2],
    )
    def distance(self) -> float:
        """Planet semi major axis from parent star (AU)."""
        return self.get_planet_semimajoraxis(unit="AU")

    @distance.setter
    def distance(self, value: float) -> None:
        """Set planet semi major axis from parent star (AU)."""
        self.set_planet_semimajoraxis(value, unit="AU")

    @fitparam(
        param_name="planet_sma",
        param_latex="$D_{planet}$",
        default_fit=False,
        default_bounds=[1, 2],
    )
    def semiMajorAxis(self) -> float:  # noqa: N802
        """Planet semi major axis from parent star (AU) (ALIAS)"""
        return self.get_planet_semimajoraxis(unit="AU")

    @semiMajorAxis.setter
    def semiMajorAxis(self, value: float) -> None:  # noqa: N802
        """Set planet semi major axis from parent star (AU) (ALIAS)"""
        self.set_planet_semimajoraxis(value, unit="AU")

    @property
    def gravity(self) -> float:
        """Surface gravity in ms-2."""
        return (G * self.get_planet_mass(unit="kg")) / (
            self.get_planet_radius(unit="m") ** 2
        )

    def gravity_at_height(self, height: float) -> float:
        """Gravity at height (m) from planet in ms-2

        Parameters
        ----------
        height: float
            Height in metres from planet surface

        Returns
        -------
        g: float
            Gravity in ms-2

        """
        return (G * self.get_planet_mass(unit="kg")) / (
            (self.get_planet_radius(unit="m") + height) ** 2
        )

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write planet information to output group."""
        planet = output.create_group("Planet")

        planet.write_string("planet_type", self.__class__.__name__)
        planet.write_scalar("planet_mass", self._mass / MJUP)
        planet.write_scalar("planet_radius", self._radius / RJUP)
        planet.write_scalar("planet_distance", self._distance / AU)
        planet.write_scalar("impact_param", self._impact)
        planet.write_scalar("orbital_period", self.orbitalPeriod)
        planet.write_scalar("albedo", self.albedo)
        planet.write_scalar("transit_time", self.transitTime)

        planet.write_scalar("mass_kg", self.mass)
        planet.write_scalar("radius_m", self.radius)
        planet.write_scalar("surface_gravity", self.gravity)
        return planet

    @derivedparam(param_name="logg", param_latex="log(g)", compute=False)
    def logg(self) -> float:
        """Surface gravity (m2/s) in log10"""
        return math.log10(self.gravity)

    def calculate_scale_properties(
        self,
        temperature: npt.NDArray[np.float64],
        pressure_levels: npt.NDArray[np.float64],
        mu: npt.NDArray[np.float64],
        length_units: t.Optional[str] = "m",
    ) -> t.Tuple[
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
    ]:
        """Solve hydrostatic for altitude, gravity and scale height of the atmosphere.

        Parameters
        ----------
        temperature: array_like
            Ttaurex.utilch layer in K

        pressure_levels: array_like
            Pressure at each layer boundary in Pa

        mu: array_like
            mean moleculer weight for each layer in kg

        Returns
        -------
        z: array
            Altitude at each layer boundary
        H: array
            scale height converted to correct length units
        g: array
            gravity converted to correct length units
        deltaz:
            dz in length units

        """

        from taurex.constants import KBOLTZ
        from taurex.util import conversion_factor

        # build the altitude profile from the bottom up
        nlayers = temperature.shape[0]
        scaleheight = np.zeros(nlayers)
        g = np.zeros(nlayers)
        z = np.zeros(nlayers + 1)
        deltaz = np.zeros(nlayers + 1)

        # surface gravity (0th layer)
        g[0] = self.gravity
        # scaleheight at the surface (0th layer)
        scaleheight[0] = (KBOLTZ * temperature[0]) / (mu[0] * g[0])
        #####
        ####
        ####

        factor = conversion_factor("m", length_units)

        for i in range(1, nlayers + 1):
            deltaz[i] = (
                (-1.0)
                * scaleheight[i - 1]
                * np.log(pressure_levels[i] / pressure_levels[i - 1])
            )
            z[i] = z[i - 1] + deltaz[i]
            if i < nlayers:
                with np.errstate(over="ignore"):
                    # gravity at the i-th layer
                    g[i] = self.gravity_at_height(z[i])
                    self.debug("G[%s] = %s", i, g[i])

                with np.errstate(divide="ignore"):
                    scaleheight[i] = (KBOLTZ * temperature[i]) / (mu[i] * g[i])

        return z * factor, scaleheight * factor, g * factor, deltaz[1:] * factor

    def compute_path_length(
        self,
        altitudes: npt.NDArray[np.float64],
        viewer: npt.NDArray[np.float64],
        tangent: npt.NDArray[np.float64],
        vector_coord_sys: t.Optional[t.Literal["cartesian"]] = "cartesian",
    ):
        """Compute path length through atmosphere."""
        from taurex.util.geometry import compute_path_length_3d

        result = compute_path_length_3d(
            self.fullRadius, altitudes, viewer, tangent, coordinates=vector_coord_sys
        )

        return result

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Return list of keywords for input file."""
        raise NotImplementedError


class Planet(BasePlanet):
    """Planet class."""

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Return list of keywords for input file."""
        return (
            "simple",
            "sphere",
        )


class Earth(Planet):
    """An implementation for earth"""

    def __init__(self):
        """Initialise earth"""
        Planet.__init__(self, 5.972e24, 6371000)

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Return list of keywords for input file."""
        return ("earth",)


class Mars(Planet):
    """An implementation for mars."""

    def __init__(self) -> None:
        """Initialise mars."""
        import astropy.units as u

        radius = (0.532 * u.R_earth).to(u.jupiterRad)
        mass = (0.107 * u.M_earth).to(u.jupiterMass)
        distance = 1.524
        Planet.__init__(mass=mass.value, radius=radius.value, distance=distance)

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        return ("mars",)



"""Two point gas profile."""
import math
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.output import OutputGroup
from taurex.util import molecule_texlabel

from .gas import Gas


class TwoPointGas(Gas):
    """Two point gas profile.

    A gas profile with two different mixing layers at the surface of the
    planet and top of the atmosphere and interpolated between the two


    """

    def __init__(
        self,
        molecule_name: t.Optional[str] = "CH4",
        mix_ratio_surface: t.Optional[float] = 1e-4,
        mix_ratio_top: t.Optional[float] = 1e-8,
    ):
        """Initialize a two point gas profile.

        Parameters
        -----------
        molecule_name : str
            Name of molecule

        mix_ratio_surface : float
            Mixing ratio of the molecule on the planet surface

        mix_ratio_top : float
            Mixing ratio of the molecule at the top of the atmosphere

        """
        super().__init__("TwoPointGas", molecule_name)
        self._mix_surface = mix_ratio_surface
        self._mix_top = mix_ratio_top
        self.add_surface_param()
        self.add_top_param()
        self._mix_profile = None

    @property
    def mixProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Mixing ratio profile.

        Returns
        -------
        mix: :obj:`array`
            Mix ratio for molecule at each layer
        """
        return self._mix_profile

    @property
    def mixRatioSurface(self) -> float:  # noqa: N802
        """Abundance on the planets surface"""
        return self._mix_surface

    @property
    def mixRatioTop(self) -> float:  # noqa: N802
        """Abundance on the top of atmosphere"""
        return self._mix_top

    @mixRatioSurface.setter
    def mixRatioSurface(self, value: float) -> None:  # noqa: N802
        """Set abundance on the planets surface"""
        self._mix_surface = value

    @mixRatioTop.setter
    def mixRatioTop(self, value: float) -> None:  # noqa: N802
        """Set abundance on the top of atmosphere"""
        self._mix_top = value

    def add_surface_param(self) -> None:
        """Add surface parameter.

        Adds fittable parameter for the surface abundance of the molecule
        with format ``[molecule]_surface``

        """
        param_name = self.molecule
        param_tex = molecule_texlabel(param_name)

        param_surface = f"{param_name}_surface"
        param_surf_tex = f"{param_tex}_surface"

        def read_surf(self):
            return self._mix_surface

        def write_surf(self, value):
            self._mix_surface = value

        read_surf.__doc__ = f"Abundance of {param_name} on the planets surface"

        fget_surf = read_surf
        fset_surf = write_surf

        bounds = [1.0e-12, 0.1]

        default_fit = False
        self.add_fittable_param(
            param_surface,
            param_surf_tex,
            fget_surf,
            fset_surf,
            "log",
            default_fit,
            bounds,
        )

    def add_top_param(self):
        """Add top parameter.

        Adds fittable parameter for the top abundance of the molecule
        with format ``[molecule]_top``

        """
        param_name = self.molecule
        param_tex = molecule_texlabel(param_name)

        param_top = f"{param_name}_top"
        param_top_tex = f"{param_tex}_top"

        def read_top(self):
            return self._mix_top

        def write_top(self, value):
            self._mix_top = value

        read_top.__doc__ = f"Abundance of {param_name} on the top of atmosphere"

        fget_top = read_top
        fset_top = write_top

        bounds = [1.0e-12, 0.1]

        default_fit = False
        self.add_fittable_param(
            param_top, param_top_tex, fget_top, fset_top, "log", default_fit, bounds
        )

    def initialize_profile(
        self,
        nlayers: t.Optional[int] = None,
        temperature_profile: t.Optional[npt.NDArray[np.float64]] = None,
        pressure_profile: t.Optional[npt.NDArray[np.float64]] = None,
        altitude_profile: t.Optional[npt.NDArray[np.float64]] = None,
    ):
        """Initialize the profile.

        Parameters
        -----------
        nlayers : int
            number of layers
        temperature_profile : np.ndarray
            temperature profile
        pressure_profile : np.ndarray
            pressure profile
        altitude_profile : np.ndarray , optional
            altitude profile, deprecated

        """
        self._mix_profile = np.zeros(nlayers)

        chem_surf = self._mix_surface
        chem_top = self._mix_top
        p_surf = pressure_profile[0]
        p_top = pressure_profile[-1]

        a = (math.log10(chem_surf) - math.log10(chem_top)) / (
            math.log10(p_surf) - math.log10(p_top)
        )

        b = math.log10(chem_surf) - a * math.log10(p_surf)

        self._mix_profile[1:-1] = 10 ** (a * np.log10(pressure_profile[1:-1]) + b)
        self._mix_profile[0] = chem_surf
        self._mix_profile[-1] = chem_top

    def write(self, output: OutputGroup) -> OutputGroup:
        gas_entry = super().write(output)
        gas_entry.write_scalar("mix_ratio_top", self.mixRatioTop)
        gas_entry.write_scalar("mix_ratio_surface", self.mixRatioSurface)
        return gas_entry

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        return (
            "twopoint",
            "2point",
        )



"""Direct imaging model."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.constants import PI

from .emission import EmissionModel

def compute_direct_image_final_flux(
    f_total, planet_radius, star_distance
):
    return f_total*(planet_radius**2)/(star_distance**2)


class DirectImageModel(EmissionModel):
    """A forward model for direct imaging of exo-planets."""

    def compute_final_flux(
        self, f_total: npt.NDArray[np.float64]
    ) -> npt.NDArray[np.float64]:
        """Compute the final flux.

        This is the emission flux that is observed at the telescope directly
        from an exo-planet.

        """
        return (
            compute_direct_image_final_flux(f_total, self._planet.fullRadius, self._star.distance * 3.08567758e16)
        )

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Input keywords for this class."""
        return (
            "direct",
            "directimage",
        )



"""Isothermal temperature profile."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.data.fittable import fitparam
from taurex.output import OutputGroup

from .tprofile import TemperatureProfile


class Isothermal(TemperatureProfile):
    """An isothermal temperature-pressure profile."""

    def __init__(self, T: t.Optional[float] = 1500) -> None:  # noqa: N803
        """Initialize isothermal class

        Parameters
        ----------

        T : float
            Isothermal Temperature to set in Kelvin
        """
        super().__init__("Isothermal")

        self._iso_temp = T

    @fitparam(
        param_name="T",
        param_latex="$T$",
        default_fit=False,
        default_bounds=[300.0, 2000.0],
    )
    def isoTemperature(self) -> float:  # noqa: N802
        """Isothermal temperature in Kelvin"""
        return self._iso_temp

    @isoTemperature.setter
    def isoTemperature(self, value: float) -> None:  # noqa: N802
        self._iso_temp = value

    @property
    def profile(self) -> npt.NDArray[np.float64]:
        """Returns an isothermal temperature profile."""
        return np.full(self.nlayers, self._iso_temp, dtype=np.float64)

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write isothermal temperature profile to output group."""
        temperature = super().write(output)
        temperature.write_scalar("T", self._iso_temp)
        return temperature

    @classmethod
    def input_keywords(cls) -> t.Tuple[str]:
        """Return all input keywords."""
        return ("isothermal",)



"""Module for wrapping MPI functions.

Most functions will do nothing if mpi4py is not present and will simply
replicate expected behaviour. This allows for MPI and non-MPI code to be
run without any changes (In theory...).
"""
import enum
import logging
import typing as t
from functools import lru_cache, wraps

import numpy as np

from .types import AnyValType

T = t.TypeVar("T", bound=AnyValType)


class Ops(str, enum.Enum):
    """Available MPI operations"""

    SUM = "sum"
    PROD = "prod"
    MAX = "max"
    MIN = "min"


def has_mpi() -> bool:
    """Checks if mpi4py is installed

    Returns
    -------
    bool:
        True if mpi4py is installed, False otherwise

    """
    try:
        import mpi4py  # noqa: F401
    except ImportError:
        return False
    return True


def convert_op(operation: Ops) -> t.Any:
    """Converts string to MPI operation."""
    from mpi4py import MPI

    return getattr(MPI, str(operation.upper()))


@lru_cache(maxsize=10)
def shared_comm() -> t.Any:
    """Return shared memory communicator.

    Returns the process id within a node. Used for shared memory.
    """
    from mpi4py import MPI

    return MPI.COMM_WORLD.Split_type(MPI.COMM_TYPE_SHARED)


@lru_cache(maxsize=10)
def nprocs() -> int:
    """Gets number of processes or returns 1 if mpi is not installed

    Returns
    -------
    int:
        Rank of process or 1 if MPI is not installed

    """
    try:
        from mpi4py import MPI
    except ImportError:
        return 1

    comm = MPI.COMM_WORLD

    return comm.Get_size()


def allgather(value: T) -> t.List[T]:
    """Gathers all values from all processes or returns value if mpi is not installed

    Parameters
    ----------
    value:
        Value to gather

    Returns
    -------
    list:
        List of gathered values or list containing value if mpi is not installed

    """
    try:
        from mpi4py import MPI
    except ImportError:
        return [value]

    comm = MPI.COMM_WORLD
    data = value
    data = comm.allgather(data)

    return data


def allreduce(value: T, op: Ops) -> T:
    """Reduces all values from all processes or returns value if mpi is not installed

    Parameters
    ----------
    value:
        Value to reduce

    op: :class:`~taurex.mpi.Ops`
        Operation to perform

    Returns
    -------
    result:
        Reduced value or value if mpi is not installed

    """

    try:
        from mpi4py import MPI
    except ImportError:
        return value

    comm = MPI.COMM_WORLD
    data = value
    data = comm.allreduce(value, op=convert_op(op))

    return data


def broadcast(array: T, rank: t.Optional[int] = 0) -> T:
    """Broadcasts array from rank or returns array if mpi is not installed

    Parameters
    ----------
    array:
        Array to broadcast

    rank: int, optional
        Rank to broadcast from, default is 0

    Returns
    -------
    array:
        Broadcasted array or array if mpi is not installed

    """
    import numpy as np

    try:
        from mpi4py import MPI
    except ImportError:
        return array
    comm = MPI.COMM_WORLD
    if isinstance(array, np.ndarray):
        data = None
        if get_rank() == rank:
            data = np.copy(array)
        else:
            data = np.zeros_like(array)
        comm.Bcast(data, root=rank)
    else:
        data = comm.bcast(array, root=rank)

    return data


@lru_cache(maxsize=10)
def get_rank(comm: t.Any = None) -> int:
    """Gets rank or returns 0 if mpi is not installed

    Parameters
    ----------

    comm: int, optional
        MPI communicator, default is MPI_COMM_WORLD


    Returns
    -------
    int:
        Rank of process in communitor or 0 if MPI is not installed

    """
    try:
        from mpi4py import MPI
    except ImportError:
        return 0

    comm = comm or MPI.COMM_WORLD
    rank = comm.Get_rank()

    return rank


def barrier(comm: t.Any = None) -> None:
    """Waits for all processes to finish.

    Does nothing if mpi4py not present

    Parameters
    ----------

    comm: int, optional
        MPI communicator, default is MPI_COMM_WORLD

    """

    try:
        from mpi4py import MPI
    except ImportError:
        return

    comm = comm or MPI.COMM_WORLD
    comm.Barrier()


def only_master_rank(f) -> t.Callable:
    """A decorator to ensure only the master MPI rank can run it."""

    @wraps(f)
    def wrapper(*args, **kwargs):
        if get_rank() == 0:
            return f(*args, **kwargs)

    return wrapper


@lru_cache(maxsize=10)
def shared_rank() -> int:
    """Gets rank within shared memory communicator. (MPI only)"""
    return shared_comm().Get_rank()


def allocate_as_shared(
    arr: np.ndarray,
    logger: t.Optional[logging.Logger] = None,
    force_shared: t.Optional[bool] = False,
):
    """Converts a numpy array into an MPI shared memory.

    This allow for things like opacities to be loaded only
    once per node when using MPI. Only activates if mpi4py
    installed and when enabled via the ``mpi_use_shared`` input::

        [Global]
        mpi_use_shared = True

    or ``force_shared=True`` otherwise does nothing and
    returns the same array back

    Parameters
    ----------

    arr: numpy array
        Array to convert

    logger: :class:`~taurex.log.logger.Logger`
        Logger object to print outputs

    force_shared: bool
        Force conversion to shared memory


    Returns
    -------
    array:
        If enabled and MPI present, shared memory version of array
        otherwise the original array

    """
    try:
        from mpi4py import MPI
    except ImportError:
        return arr
    from taurex.cache import GlobalCache

    if GlobalCache()["mpi_use_shared"] or force_shared:
        if logger is not None:
            logger.info("Moving to shared memory")
        comm = shared_comm()
        nbytes = arr.size * arr.itemsize

        window = MPI.Win.Allocate_shared(nbytes, arr.itemsize, comm=comm)
        buf, itemsize = window.Shared_query(0)
        if itemsize != arr.itemsize:
            raise Exception(
                f"Shared memory size {itemsize} != array itemsize {arr.itemsize}"
            )

        shared_array = np.ndarray(buffer=buf, dtype=arr.dtype, shape=arr.shape)

        if shared_rank() == 0:
            np.copyto(shared_array, arr)

        comm.Barrier()

        return shared_array
    else:
        return arr



"""Alias for pressure profiles."""

from taurex.data.profiles.pressure import (
    ArrayPressureProfile,
    FilePressureProfile,
    PressureProfile,
    SimplePressureProfile,
)

__all__ = [
    "PressureProfile",
    "SimplePressureProfile",
    "FilePressureProfile",
    "ArrayPressureProfile",
]



"""Module to handle outputs and storage from various models."""
from .hdf5 import HDF5Output
from .output import Output, OutputGroup
from .writeable import Writeable

__all__ = ["Output", "HDF5Output", "Writeable", "OutputGroup"]



"""Handling of molecular absorption."""
import math
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.cache import GlobalCache, OpacityCache
from taurex.cache.ktablecache import KTableCache
from taurex.model.model import ForwardModel

from .contribution import Contribution

contribute_ktau: t.Callable[
    [
        int,
        int,
        int,
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        int,
        int,
        int,
    ],
    npt.NDArray[np.float64],
] = None


def contribute_ktau_numba(
    startk: int,
    endk: int,
    density_offset: int,
    sigma: npt.NDArray[np.float64],
    density: npt.NDArray[np.float64],
    path: npt.NDArray[np.float64],
    weights: npt.NDArray[np.float64],
    tau: npt.NDArray[np.float64],
    ngrid: int,
    layer: int,
    ngauss: int,
) -> npt.NDArray[np.float64]:
    """Integrate the optical depth for a given layer.

    This version uses numba to speed up the calculation.

    Parameters
    ----------
    startk : int
        Starting integration index of the layer
    endk : int
        Ending integration index of the layer
    density_offset : int
        Offset of the density profile
    sigma : npt.NDArray[np.float64]
        Cross-sections
    density : npt.NDArray[np.float64]
        Density profile
    path : npt.NDArray[np.float64]
        Path length
    weights : npt.NDArray[np.float64]
        Weights for the gauss quadrature
    tau : npt.NDArray[np.float64]
        Optical depth
    ngrid : int
        Number of grid points
    layer : int
        Layer index
    ngauss : int
        Number of gauss points

    """
    tau_temp = np.zeros(shape=(ngrid, ngauss))

    for k in range(startk, endk):
        _path = path[k]
        _density = density[k + density_offset]
        # for mol in range(nmols):
        for wn in range(ngrid):
            for g in range(ngauss):
                tau_temp[wn, g] += sigma[k + layer, wn, g] * _path * _density

    for wn in range(ngrid):
        transtemp = 0.0
        for g in range(ngauss):
            transtemp += math.exp(-tau_temp[wn, g]) * weights[g]
        tau[layer, wn] += -math.log(transtemp)


def contribute_ktau_numpy(
    startk: int,
    endk: int,
    density_offset: int,
    sigma: npt.NDArray[np.float64],
    density: npt.NDArray[np.float64],
    path: npt.NDArray[np.float64],
    weights: npt.NDArray[np.float64],
    tau: npt.NDArray[np.float64],
    ngrid: int,  # For compatibility
    layer: int,
    ngauss: int,  # For compatibility
) -> npt.NDArray[np.float64]:
    """Integrate the optical depth for a given layer with numpy.

    This version uses numba to speed up the calculation.

    Parameters
    ----------
    startk : int
        Starting integration index of the layer
    endk : int
        Ending integration index of the layer
    density_offset : int
        Offset of the density profile
    sigma : npt.NDArray[np.float64]
        Cross-sections
    density : npt.NDArray[np.float64]
        Density profile
    path : npt.NDArray[np.float64]
        Path length
    weights : npt.NDArray[np.float64]
        Weights for the gauss quadrature
    tau : npt.NDArray[np.float64]
        Optical depth
    ngrid : int
        Number of grid points
    layer : int
        Layer index
    ngauss : int
        Number of gauss points

    """
    _path = path[startk:endk, None, None]
    _density = density[startk + density_offset : endk + density_offset]
    _sigma = sigma[layer + startk : layer + endk]
    tau_temp = np.sum(
        _sigma * _path[..., None, None] * _density[..., None, None],
        axis=0,
    )

    transtemp = np.sum(np.exp(-tau_temp) * weights[None, :], axis=-1)

    tau[layer] += -np.log(transtemp)

    return tau


try:
    import numba

    contribute_ktau = numba.jit(
        contribute_ktau_numba, nopython=True, nogil=True, fastmath=True
    )
except ImportError:
    contribute_ktau = contribute_ktau_numpy


class AbsorptionContribution(Contribution):
    """Absorption contribution."""

    def contribute(
        self,
        model: ForwardModel,
        start_horz_layer: int,
        end_horz_layer: int,
        density_offset: int,
        layer: int,
        density: npt.NDArray[np.float64],
        tau: npt.NDArray[np.float64],
        path_length: t.Optional[npt.NDArray[np.float64]] = None,
    ) -> None:
        """Contribute to the optical depth.

        Parameters
        ----------
        model : ForwardModel
            Forward model
        start_horz_layer : int
            Starting horizontal layer
        end_horz_layer : int
            Ending horizontal layer
        density_offset : int
            Offset of the density profile
        layer : int
            Layer index
        density : npt.NDArray[np.float64]
            Density profile
        tau : npt.NDArray[np.float64]
            Optical depth
        path_length : npt.NDArray[np.float64], optional
            Path length, by default None
        """
        if self._use_ktables:
            # startK,endK,density_offset,sigma,density,path,weights,tau,ngrid,layer,ngauss

            contribute_ktau(
                start_horz_layer,
                end_horz_layer,
                density_offset,
                self.sigma_xsec,
                density,
                path_length,
                self.weights,
                tau,
                self._ngrid,
                layer,
                self.weights.shape[0],
            )
        else:
            super().contribute(
                model,
                start_horz_layer,
                end_horz_layer,
                density_offset,
                layer,
                density,
                tau,
                path_length,
            )

    def __init__(self) -> None:
        super().__init__("Absorption")
        self._opacity_cache = OpacityCache()

    def prepare_each(
        self, model: ForwardModel, wngrid: npt.NDArray[np.float64]
    ) -> t.Generator[t.Tuple[str, npt.NDArray[np.float64]], None, None]:
        """Prepare each component opacity.

        Parameters
        ----------
        model : ForwardModel
            Forward model

        wngrid : npt.NDArray[np.float64]
            Wavenumber grid

        Yields
        ------
        t.Generator[t.Tuple[str, npt.NDArray[np.float64]], None, None]
            Gas name and opacity

        """
        self.debug("Preparing model with %s", wngrid.shape)
        self._ngrid = wngrid.shape[0]
        self._use_ktables = GlobalCache()["opacity_method"] == "ktables"
        self.info("Using cross-sections? %s", not self._use_ktables)

        if self._use_ktables:
            self._opacity_cache = KTableCache()
        else:
            self._opacity_cache = OpacityCache()
        sigma_xsec = None
        self.weights = None

        for gas in model.chemistry.activeGases:
            # self._total_contrib[...] =0.0
            gas_mix = model.chemistry.get_gas_mix_profile(gas)
            self.info("Recomputing active gas %s opacity", gas)

            xsec = self._opacity_cache[gas]

            if self._use_ktables and self.weights is None:
                self.weights = xsec.weights

            if sigma_xsec is None:
                if self._use_ktables:
                    sigma_xsec = np.zeros(
                        shape=(self._nlayers, self._ngrid, len(self.weights))
                    )
                else:
                    sigma_xsec = np.zeros(shape=(self._nlayers, self._ngrid))
            else:
                sigma_xsec[...] = 0.0

            for idx_layer, tp in enumerate(
                zip(model.temperatureProfile, model.pressureProfile)
            ):
                self.debug("Got index,tp %s %s", idx_layer, tp)

                temperature, pressure = tp
                # print(gas,self._opacity_cache[gas].opacity(temperature,pressure,wngrid),gas_mix[idx_layer])
                sigma_xsec[idx_layer] += (
                    xsec.opacity(temperature, pressure, wngrid) * gas_mix[idx_layer]
                )

            self.sigma_xsec = sigma_xsec

            self.debug("SIGMAXSEC %s", self.sigma_xsec)

            yield gas, sigma_xsec

    def prepare(self, model: ForwardModel, wngrid: npt.NDArray[np.float64]) -> None:
        """Used to prepare the contribution for the calculation.
        Called before the forward model performs the main optical depth
        calculation. Default behaviour is to loop through :func:`prepare_each`
        and sum all results into a single cross-section.

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        wngrid: :obj:`array`
            Wavenumber grid
        """

        self._ngrid = wngrid.shape[0]
        self._nlayers = model.nLayers

        sigma_xsec = None
        self.debug("ABSORPTION VERSION")
        for gas, sigma in self.prepare_each(model, wngrid):
            self.debug("Gas %s", gas)
            self.debug("Sigma %s", sigma)
            if sigma_xsec is None:
                sigma_xsec = np.zeros_like(sigma)
            sigma_xsec += sigma

        self.sigma_xsec = sigma_xsec
        self.debug("Final sigma is %s", self.sigma_xsec)
        self.info("Done")

    def finalize(self, model: ForwardModel) -> None:
        """Finalize the contribution."""
        raise NotImplementedError

    @property
    def sigma(self) -> npt.NDArray[np.float64]:
        """Return effective cross-sections."""
        return self.sigma_xsec

    @classmethod
    def input_keywords(cls) -> t.Tuple[str]:
        """Return input keywords for the contribution."""
        return (
            "Absorption",
            "Molecules",
        )



"""Module for correlated-k tables opacities."""
from .hdfktable import HDF5KTable
from .ktable import KTable
from .nemesisktables import NemesisKTables
from .picklektable import PickleKTable  # noqa: S403

__all__ = ["KTable", "HDF5KTable", "NemesisKTables", "PickleKTable"]



"""Module for handling loading models from HDF5 files."""

import numpy as np

from ..parameter.factory import get_keywordarg_dict
from .util import class_for_name


def get_klass_args(klass):
    return list(get_keywordarg_dict(klass)[0].keys())


def load_generic_profile_from_hdf5(
    loc, module, identifier, profile_type=None, premade_dict=None, replacement_dict=None
):
    if profile_type is None:
        profile_type = loc[identifier][()]

    temp_keys = list(loc.keys())
    klass = class_for_name(profile_type)

    klass_kwargs = get_klass_args(klass)

    args_dict = premade_dict or {}
    repl_dict = replacement_dict or {}

    for kw in klass_kwargs:
        if kw in temp_keys:
            v = loc[kw][()]
            if isinstance(v, np.ndarray) and v.dtype.type is np.string_:
                from taurex.util import decode_string_array

                v = decode_string_array(v)
            if kw in repl_dict:
                args_dict[kw] = repl_dict[kw]
            else:
                try:
                    v = v.decode()
                except (
                    AttributeError,
                    UnicodeDecodeError,
                ):
                    pass
                args_dict[kw] = v

    return klass(**args_dict)


def load_temperature_from_hdf5(loc, replacement_dict=None):
    return load_generic_profile_from_hdf5(
        loc["Temperature"],
        "taurex.data.profiles.temperature",
        "temperature_type",
        replacement_dict=replacement_dict,
    )


def load_pressure_from_hdf5(loc, replacement_dict=None):
    return load_generic_profile_from_hdf5(
        loc["Pressure"],
        "taurex.data.profiles.pressure",
        "pressure_type",
        replacement_dict=replacement_dict,
    )


def load_gas_from_hdf5(loc, molecule, replacement_dict=None):
    return load_generic_profile_from_hdf5(
        loc[molecule],
        "taurex.data.profiles.chemistry",
        "gas_type",
        replacement_dict=replacement_dict,
    )


def load_planet_from_hdf5(loc, replacement_dict=None):
    return load_generic_profile_from_hdf5(
        loc["Planet"],
        "taurex.data.planet",
        "planet_type",
        profile_type="Planet",
        replacement_dict=replacement_dict,
    )


def load_star_from_hdf5(loc, replacement_dict=None):
    return load_generic_profile_from_hdf5(
        loc["Star"],
        "taurex.data.stellar",
        "star_type",
        replacement_dict=replacement_dict,
    )


def load_chemistry_from_hdf5(loc, replacement_dict=None):
    from taurex.data.profiles.chemistry import TaurexChemistry
    from taurex.util import decode_string_array

    chemistry = load_generic_profile_from_hdf5(
        loc["Chemistry"],
        "taurex.data.profiles.chemistry",
        "chemistry_type",
        replacement_dict=replacement_dict,
    )

    if isinstance(chemistry, TaurexChemistry):
        for mol in decode_string_array(loc["Chemistry"]["active_gases"][()]):
            if mol not in chemistry._fill_gases:
                chemistry.addGas(load_gas_from_hdf5(loc["Chemistry"], mol))
        for mol in decode_string_array(loc["Chemistry"]["inactive_gases"]):
            if mol not in chemistry._fill_gases:
                chemistry.addGas(load_gas_from_hdf5(loc["Chemistry"], mol))

    return chemistry


def load_contrib_from_hdf5(loc, contribution, replacement_dict=None):
    return load_generic_profile_from_hdf5(
        loc[contribution],
        "taurex.contributions",
        "contrib_type",
        profile_type=contribution,
        replacement_dict=replacement_dict,
    )


def load_model_from_hdf5(loc, replacement_dict=None):
    chemistry = load_chemistry_from_hdf5(loc, replacement_dict=replacement_dict)
    pressure = load_pressure_from_hdf5(loc, replacement_dict=replacement_dict)
    temperature = load_temperature_from_hdf5(loc, replacement_dict=replacement_dict)
    planet = load_planet_from_hdf5(loc, replacement_dict=replacement_dict)
    star = load_star_from_hdf5(loc, replacement_dict=replacement_dict)

    kwargs = {}

    kwargs["planet"] = planet
    kwargs["star"] = star
    kwargs["chemistry"] = chemistry
    kwargs["temperature_profile"] = temperature
    kwargs["pressure_profile"] = pressure

    model = load_generic_profile_from_hdf5(
        loc,
        "taurex.model",
        "model_type",
        premade_dict=kwargs,
        replacement_dict=replacement_dict,
    )

    contribution_loc = loc["Contributions"]

    def contrib_iterator(loc):
        import h5py

        for key in loc.keys():
            item = loc[key]
            if isinstance(item, h5py.Group):
                yield key

    for contrib in contrib_iterator(contribution_loc):
        model.add_contribution(
            load_contrib_from_hdf5(
                contribution_loc, contrib, replacement_dict=replacement_dict
            )
        )

    return model


def taurex_hdf5_to_model(filename, replacement_dict=None):
    import h5py

    with h5py.File(filename, "r") as f:
        model = load_model_from_hdf5(
            f["ModelParameters"], replacement_dict=replacement_dict
        )

    return model


def taurex_hdf5_to_observation(filename):
    import h5py

    from taurex.data.spectrum import ArraySpectrum
    from taurex.util import wnwidth_to_wlwidth

    with h5py.File(filename, "r") as f:
        try:
            instrument_section = f["Output"]["Spectra"]
        except KeyError as e:
            raise KeyError(
                "No instrument data found in HDF5 or retrieval hdf5 used"
            ) from e
        inst_wngrid = instrument_section["instrument_wngrid"][...]
        inst_spectrum = instrument_section["instrument_spectrum"][...]
        inst_noise = instrument_section["instrument_noise"][...]
        inst_width = instrument_section["instrument_wnwidth"][...]

        inst_wlgrid = 10000 / inst_wngrid

        inst_wlwidth = wnwidth_to_wlwidth(inst_wngrid, inst_width)
        observation = ArraySpectrum(
            np.vstack([inst_wlgrid, inst_spectrum, inst_noise, inst_wlwidth]).T
        )
    return observation



"""Contains caching class for Molecular ktable files."""

import typing as t

from taurex.log import Logger

from ..opacity.ktables import KTable
from . import GlobalCache
from .singleton import Singleton


class KTableCache(Singleton):
    """Implements a lazy load of opacities.

    A singleton that loads and caches ktables as they are needed.

    Now TauREx3 will use it instead in all calculations!

    """

    def init(self) -> None:
        """Initialise the cache."""
        self.opacity_dict: t.Dict[str, KTable] = {}
        self._opacity_path = GlobalCache()["ktable_path"]
        self.log = Logger("KTableCache")
        self._default_interpolation = "linear"
        self._memory_mode = True

    def set_ktable_path(self, opacity_path: str) -> None:
        """Set the path(s) that will be searched for ktables.

        Opacities in this path must be of supported types:

            - HDF5 opacities
            - ``.pickle`` opacities
            - ExoTransmit opacities.

        Parameters
        ----------

        opacity_path : str or :obj:`list` of str, optional
            search path(s) to look for molecular opacities

        """

        import os

        GlobalCache()["ktable_path"] = opacity_path

        if not os.path.isdir(opacity_path):
            self.log.error("PATH: %s does not exist!!!", opacity_path)
            raise NotADirectoryError
        self.log.debug("Path set to %s", opacity_path)

    def __getitem__(self, key: str) -> KTable:
        """For a molecule return the relevant opacity.

        Opacity return is an :class:`~taurex.opacity.opacity.Opacity` object.


        Parameter
        ---------
        key : str
            molecule name

        Returns
        -------
        :class:`~taurex.opacity.pickleopacity.PickleOpacity`
            Cross-section object desired

        Raises
        ------
        Exception
            If molecule could not be loaded/found

        """
        if key in self.opacity_dict:
            return self.opacity_dict[key]
        else:
            # Try a load of the opacity
            self.load_opacity(molecule_filter=[key])
            # If we have it after a load then good job boys
            if key in self.opacity_dict:
                return self.opacity_dict[key]
            else:
                # try:
                #     # if self._radis:
                #     #     return self.create_radis_opacity(key,molecule_filter=[key])
                #     # else:
                #         raise Exception
                # except Exception as e:
                # Otherwise throw an error
                self.log.error("Opacity for molecule %s could not be loaded", key)
                self.log.error(
                    "It could not be found in the local dictionary %s",
                    list(self.opacity_dict.keys()),
                )
                self.log.error("Or paths %s", self._opacity_path)
                self.log.error("Try loading it manually/ putting it in a path")
                raise Exception("Opacity could not be loaded")

    def add_opacity(
        self, opacity: KTable, molecule_filter: t.Optional[t.List[str]] = None
    ) -> None:
        """Adds a :class:`~taurex.opacity.ktable.KTable` object to the cache

        Parameters
        ----------
        opacity : :class:`~taurex.opacity.ktable.KTable`
            Opacity object to add to the cache

        molecule_filter : :obj:`list` of str , optional
            If provided, the opacity object will only be included
            if its molecule is in the list. Mostly used by the
            :func:`__getitem__` for filtering

        """
        self.log.info("Reading opacity %s", opacity.moleculeName)
        if opacity.moleculeName in self.opacity_dict:
            self.log.warning(
                "Opacity with name %s already in opactiy dictionary %s skipping",
                opacity.moleculeName,
                self.opacity_dict.keys(),
            )
            return
        if molecule_filter is not None:
            if opacity.moleculeName in molecule_filter:
                self.log.info("Loading opacity %s into model", opacity.moleculeName)
                self.opacity_dict[opacity.moleculeName] = opacity
        else:
            self.log.info("Loading opacity %s into model", opacity.moleculeName)
            self.opacity_dict[opacity.moleculeName] = opacity

    def find_list_of_molecules(self) -> t.Set[str]:
        """Returns a list of molecules that can be loaded.

        Returns
        -------
        :obj:`set` of str
            List of molecules that can be loaded

        """
        from taurex.parameter.classfactory import ClassFactory

        opacity_klasses = ClassFactory().ktableKlasses

        molecules = []

        for c in opacity_klasses:
            molecules.extend([x[0] for x in c.discover()])

        forced = []

        return set(molecules + forced)

    def load_opacity_from_path(
        self, path: str, molecule_filter: t.Optional[t.List[str]] = None
    ) -> None:
        """Searches path for and loads molecular cross-section files.

        For example, if the path is ``/path/to/opacities`` then
        ``/path/to/opacities/H2O.pickle`` will be loaded as
        :class:`~taurex.opacity.pickleopacity.PickleOpacity`


        Parameters
        ----------
        path : str
            Path to search for molecular cross-section files

        molecule_filter : :obj:`list` of str , optional
            If provided, the opacity will only be loaded
            if its molecule is in this list. Mostly used by the
            :func:`__getitem__` for filtering

        """

        from taurex.parameter.classfactory import ClassFactory

        cf = ClassFactory()

        opacity_klass_list = sorted(cf.ktableKlasses, key=lambda x: x.priority())
        for c in opacity_klass_list:
            try:
                discover = c.discover()
            except NotImplementedError:
                self.log.warning("Klass %s has no discover method", c.__name__)
                continue
            for mol, args in discover:
                self.log.debug("Klass: %s %s", mol, args)
                op = None
                if mol in molecule_filter:
                    if not isinstance(
                        args,
                        (
                            list,
                            tuple,
                        ),
                    ):
                        args = [args]
                    op = c(*args)
                if op is not None and op.moleculeName not in self.opacity_dict:
                    self.add_opacity(op, molecule_filter=molecule_filter)

    def load_opacity(
        self,
        opacities: t.Optional[t.Union[KTable, t.List[KTable]]] = None,
        opacity_path: t.Optional[t.Union[str, t.List[str]]] = None,
        molecule_filter: t.Optional[t.Union[str, t.List[str]]] = None,
    ) -> None:
        """Main function to use when loading molecular opacities.

        Handles both cross sections and paths. Handles lists of either so lists of
        :class:`~taurex.opacity.opacity.Opacity` objects or lists of paths can be used
        to load multiple files/objects


        Parameters
        ----------
        opacities : :class:`~taurex.opacity.opacity.Opacity` or
        :obj:`list` of :class:`~taurex.opacity.opacity.Opacity` , optional
            Object(s) to include in cache

        opacity_path : str or :obj:`list` of str, optional
            search path(s) to look for molecular opacities

        molecule_filter : :obj:`list` of str , optional
            If provided, the opacity will only be loaded
            if its molecule is in this list. Mostly used by the
            :func:`__getitem__` for filtering

        """
        from taurex.opacity import Opacity

        if opacity_path is None:
            opacity_path = self._opacity_path

        if opacities is not None:
            if isinstance(opacities, (list,)):
                self.log.debug("Opacity passed is list")
                for opacity in opacities:
                    self.add_opacity(opacity, molecule_filter=molecule_filter)
            elif isinstance(opacities, Opacity):
                self.add_opacity(opacities, molecule_filter=molecule_filter)
            else:
                self.log.error(
                    "Unknown type %s passed into opacities, should be a list, single \
                     opacity or None if reading a path",
                    type(opacities),
                )
                raise Exception("Unknown type passed into opacities")
        else:
            self.load_opacity_from_path(opacity_path, molecule_filter=molecule_filter)
            # if isinstance(opacity_path,(list,)):
            #     for path in opacity_path:
            #         self.load_opacity_from_path(path,molecule_filter=molecule_filter)

    def clear_cache(self) -> None:
        """Clears all currently loaded cross-sections."""
        self.opacity_dict = {}
        self._opacity_path = GlobalCache()["ktable_path"]



"""Chemistry from file module."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.output import OutputGroup
from taurex.types import PathLike

from .autochemistry import AutoChemistry


class ChemistryFile(AutoChemistry):
    """Chemistry profile read from file."""

    def __init__(self, gases: t.List[str] = None, filename: PathLike = None):
        """Initialize chemistry from file.


        Parameters
        ----------

        gases : :obj:`list`
            Gases in file

        filename : PathLike
            filename for mix ratios

        """
        super().__init__(self.__class__.__name__)

        self._gases = gases
        self._filename = filename
        self._mix_ratios = np.loadtxt(filename).T
        self.determine_active_inactive()

    @property
    def gases(self) -> t.List[str]:
        """Gases in file."""
        return self._gases

    @property
    def mixProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Mix ratios."""
        return self._mix_ratios

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write chemistry to output group."""
        gas_entry = super().write(output)
        gas_entry.write_scalar("filename", self._filename)

        return gas_entry

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Input keywords for chemistry from file."""
        return (
            "file",
            "fromfile",
        )



import typing as t


class MalformedPriorInputError(Exception):
    """Raised when the input prior string is malformed"""

    pass


def validate_priors(prior_string: str) -> None:
    count_left_bracket = prior_string.count("(")
    count_right_bracket = prior_string.count(")")

    if (
        count_left_bracket != count_right_bracket
        or count_left_bracket == 0
        or count_right_bracket == 0
    ):
        raise MalformedPriorInputError("Parenthesis are not balanced")

    # if count_left_bracket != 1:
    #     raise MalformedPriorInput("Only a single parenthesis pair allowed")


def parse_priors(prior_string: str) -> t.Tuple[str, t.Dict[str, t.Any]]:
    import ast

    func_parse = ast.parse(prior_string)

    actual_func = func_parse.body[0].value

    function_name = actual_func.func.id

    func_args = {kw.arg: ast.literal_eval(kw.value) for kw in actual_func.keywords}

    return function_name, func_args



"""Class for handling ktables from NEMESIS."""
import pathlib
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.types import PathLike
from taurex.util import sanitize_molecule_string

from ..interpolateopacity import InterpModeType, InterpolatingOpacity
from .ktable import KTable


class NemesisKTables(KTable, InterpolatingOpacity):
    """KTables from NEMESIS."""

    @classmethod
    def discover(
        cls,
    ) -> t.Optional[t.List[t.Tuple[str, t.Tuple[pathlib.Path, InterpModeType]]]]:
        """Discover NEMESIS ktables in path."""
        import pathlib

        from taurex.cache import GlobalCache

        path = GlobalCache()["ktable_path"]
        if path is None:
            return []
        path = pathlib.Path(path)

        files = path.glob("*.kta")

        discovery = []

        interp = GlobalCache()["xsec_interpolation"] or "linear"

        for f in files:
            splits = pathlib.Path(f).stem.split("_")
            mol_name = sanitize_molecule_string(splits[0])

            discovery.append((mol_name, [f, interp]))

        return discovery

    def __init__(
        self,
        filename: PathLike,
        interpolation_mode: t.Optional[InterpModeType] = "linear",
    ):
        """Initialize and read NEMESIS ktable.

        Parameters
        ----------
        filename : PathLike
            Path to file.
        interpolation_mode : t.Optional[InterpModeType], optional
            Interpolation mode, by default "linear".


        """
        filename = pathlib.Path(filename)
        super().__init__(
            f"NemesisKtable:{filename.stem[0:10]}",
            interpolation_mode=interpolation_mode,
        )
        if not filename.exists():
            raise FileNotFoundError(f"Could not find file {filename}")
        self._filename = filename
        splits = filename.stem.split("_")
        mol_name = sanitize_molecule_string(splits[0])
        self._molecule_name = mol_name
        self._spec_dict = None
        self._resolution = None
        self._decode_ktables(filename)

    @property
    def moleculeName(self) -> str:  # noqa: N802
        """Molecule name."""
        return self._molecule_name

    @property
    def xsecGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Cross section grid."""
        return self._xsec_grid

    def _decode_ktables(self, filename: pathlib.Path):
        """Decode NEMESIS ktable format."""
        self.debug("Reading NEMESIS FORMAT")
        nem_file_float = np.fromfile(filename, dtype=np.float32)
        nem_file_int = nem_file_float.view(np.int32)
        array_counter = 0
        self.debug("MAGIC NUMBER: %s", nem_file_int[0])
        wncount = nem_file_int[1]
        self.debug("WNCOUNT = %s", wncount)
        wnstart = nem_file_float[2]
        self.debug("WNSTART = %s um", wnstart)
        float_num = nem_file_float[3]
        self.debug("FLOAT: %s INT: %s", float_num, nem_file_int[4])

        num_pressure = nem_file_int[5]
        num_temperature = nem_file_int[6]
        num_quads = nem_file_int[7]

        self.debug("NP: %s NT: %s NQ: %s", num_pressure, num_temperature, num_quads)
        self.debug("UNKNOWN VALUES: %s %s", nem_file_int[8], nem_file_int[9])

        array_counter += 10 + num_quads * 2
        self._samples, self._weights = (
            nem_file_float[10:array_counter].reshape(2, -1).astype(np.float64)
        )
        self.debug("Samples: %s, Weights: %s", self._samples, self._weights)
        self.debug("%s", nem_file_int[array_counter])
        array_counter += 1
        self.debug("%s", nem_file_int[array_counter])
        array_counter += 1
        self._pressure_grid = (
            nem_file_float[array_counter : array_counter + num_pressure].astype(
                np.float64
            )
            * 1e5
        )
        self.debug("Pgrid: %s", self._pressure_grid)
        array_counter += num_pressure
        self._temperature_grid = nem_file_float[
            array_counter : array_counter + num_temperature
        ].astype(np.float64)
        array_counter += num_temperature
        self.debug("Tgrid: %s", self._temperature_grid)
        self._wavenumber_grid = 10000 / nem_file_float[
            array_counter : array_counter + wncount
        ].astype(np.float64)
        self._wavenumber_grid = self._wavenumber_grid[::-1]
        array_counter += wncount
        self.debug("Wngrid: %s", self._wavenumber_grid)

        self._xsec_grid = (
            nem_file_float[array_counter:].reshape(
                wncount, num_pressure, num_temperature, num_quads
            )
            * 1e-20
        ).astype(np.float64)
        self._xsec_grid = self._xsec_grid.transpose((1, 2, 0, 3))
        self._xsec_grid = self._xsec_grid[::, ::, ::-1, :]
        self._min_pressure = self._pressure_grid.min()
        self._max_pressure = self._pressure_grid.max()
        self._min_temperature = self._temperature_grid.min()
        self._max_temperature = self._temperature_grid.max()

    @property
    def wavenumberGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Wavenumber grid."""
        return self._wavenumber_grid

    @property
    def temperatureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Temperature grid."""
        return self._temperature_grid

    @property
    def pressureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Pressure grid."""
        return self._pressure_grid

    @property
    def resolution(self) -> float:
        """Resolution of grid."""
        return self._resolution

    @property
    def weights(self) -> npt.NDArray[np.float64]:
        """Quandrature Weights."""
        return self._weights



from .optimizer import Optimizer
from .nestle import NestleOptimizer

try:
    from .multinest import MultiNestOptimizer
except ImportError:
    pass

try:
    from .polychord import PolyChordOptimizer
except ImportError:
    pass

try:
    from .dypolychord import dyPolyChordOptimizer
except ImportError:
    pass


"""Instrument implementation for SNR noise model."""
import math
import typing as t

import numpy as np

from taurex.binning import BinDownType, Binner
from taurex.model import ForwardModel
from taurex.types import ModelOutputType

from .instrument import Instrument


class SNRInstrument(Instrument):
    """SNR noise model.

    Simple instrument model that, for a given
    wavelength-independant, signal-to-noise ratio,
    compute resulting noise from it.

    """

    def __init__(
        self, SNR: t.Optional[int] = 10, binner: t.Optional[Binner] = None  # noqa: N803
    ):
        """Initialize SNR instrument.

        Parameters
        ----------

        SNR: float
            Signal-to-noise ratio

        binner: :class:`~taurex.binning.binner.Binner`, optional
            Optional resampler to generate a new spectral
            grid.


        """
        super().__init__()

        self._binner = binner
        self._SNR = SNR

    def model_noise(
        self,
        model: ForwardModel,
        model_res: t.Optional[ModelOutputType] = None,
        num_observations: t.Optional[int] = 1,
    ) -> BinDownType:
        """Attach noise to forward model.

        Parameters
        ----------

        model:
            Forward model to pass.

        model_res:
            Result from :func:`~taurex.model.model.ForwardModel.model`

        num_observations:
            Number of observations to simulate

        """
        if model_res is None:
            model_res = model.model()

        binner = self._binner
        if binner is None:
            binner = model.defaultBinner()

        wngrid, spectrum, error, grid_width = self._binner.bin_model(model_res)

        signal = spectrum.max() - spectrum.min()

        noise = np.ones(spectrum.shape) * signal / self._SNR

        return wngrid, spectrum, noise / math.sqrt(num_observations), grid_width

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        return (
            "snr",
            "SNR",
            "signal-noise-ratio",
        )



"""Module for absorption opacity."""
from .exotransmit import ExoTransmitOpacity
from .hdf5opacity import HDF5Opacity
from .interpolateopacity import InterpModeType, InterpolatingOpacity
from .opacity import Opacity
from .pickleopacity import PickleOpacity  # noqa

__all__ = [
    "PickleOpacity",
    "InterpolatingOpacity",
    "HDF5Opacity",
    "ExoTransmitOpacity",
    "Opacity",
    "InterpModeType",
]



"""Handles citation information for Taurex."""
import typing as t
from functools import lru_cache
from urllib.error import HTTPError, URLError

from taurex.log.logger import setup_log

_log = setup_log(__name__)

try:
    from pybtex.database import Entry

    has_pybtex = True
except ImportError:
    _log.warning("Pybtex not installed. Citation functionality disabled")
    has_pybtex = False
    Entry = None


def cleanup_string(string: str) -> str:
    """Cleans up a string for bibtex."""
    return string.replace("{", "").replace("}", "").replace("\\", "")


def recurse_bibtex(
    obj: "Citable", entries: t.Optional[t.List[str]] = None
) -> t.List[str]:
    """Recursively search for bibtex entries in base classes."""
    entries = entries or []

    for b in obj.__class__.__bases__:
        if issubclass(b, Citable):
            entries.extend(b.BIBTEX_ENTRIES)
            recurse_bibtex(b, entries)
    return entries


def stringify_people(authors: t.List[str]) -> str:
    """Converts authors to a string."""
    return ", ".join([cleanup_string(str(p)) for p in authors])


def unique_citations_only(citations: t.List[str]) -> t.List[str]:
    """Removes duplicate citations."""
    current_citations = []
    for c in citations:
        if c not in current_citations:
            current_citations.append(c)
    return current_citations


def to_bibtex(citations: t.List[str]) -> str:
    """Converts citations to bibtex."""
    import uuid

    try:
        from pybtex.database import BibliographyData
    except ImportError:
        return str(citations)

    entries = {str(uuid.uuid4())[:8]: b for b in citations}
    bib_data = BibliographyData(entries=entries)

    return bib_data.to_string("bibtex")


def handle_publication(fields: t.Dict[str, str]) -> str:
    """Handles publication information."""
    journal = []
    if "journal" in fields:
        journal.append(cleanup_string(fields["journal"]))
    elif "booktitle" in fields:
        journal.append(cleanup_string(fields["booktitle"]))
    elif "archivePrefix" in fields:
        journal.append(cleanup_string(fields["archivePrefix"]))

    if "volume" in fields:
        journal.append(cleanup_string(fields["volume"]))
    elif "eprint" in fields:
        journal.append(cleanup_string(fields["eprint"]))
    if "pages" in fields:
        journal.append(cleanup_string(fields["pages"]))

    if "month" in fields:
        journal.append(cleanup_string(fields["month"]))

    if "year" in fields:
        journal.append(cleanup_string(fields["year"]))

    return ", ".join(journal)


def construct_nice_printable_string(entry: str, indent: int = 0) -> str:
    """Constructs a nice printable string for each entry."""
    mystring = ""
    indent = "".join(["\t"] * indent)
    form = f"{indent}%s\n"

    if isinstance(entry, str) or not has_pybtex:
        return f"Found non bibtex citation or pybtex not installed: {entry}\n"

    if "title" in entry.fields:
        mystring += form % cleanup_string(entry.fields["title"])

    people = entry.persons
    if "author" in people:
        mystring += form % stringify_people(people["author"])

    mystring += form % handle_publication(entry.fields)

    return mystring


class Citable:
    """Defines a class that contains citation information."""

    BIBTEX_ENTRIES = []
    """List of bibtex entries."""

    def citations(self) -> t.List[str]:
        """Returns a list of citations."""
        entries = self.BIBTEX_ENTRIES[:]
        entries = recurse_bibtex(self, entries)

        if not has_pybtex:
            return unique_citations_only(entries)
        all_citations = [Entry.from_string(b, "bibtex") for b in entries]

        return unique_citations_only(all_citations)

    def nice_citation(
        self,
        prefix: t.Optional[str] = "",
        start_idx: t.Optional[int] = 0,
        indent: t.Optional[int] = 0,
    ) -> str:
        """Returns a nice printable string of citations.

        Parameters
        ----------
        prefix: str, optional
            Prefix to add to each citation
        start_idx: int, optional
            Starting index of citation
        indent: int, optional
            Indentation of citation


        """
        entries = self.citations()

        if len(entries) == 0:
            return ""

        return "\n".join([construct_nice_printable_string(e) for e in entries])


@lru_cache(maxsize=100)
def doi_to_bibtex(doi: str) -> t.Optional[str]:
    """Converts a doi to bibtex.

    Parameters
    ----------
    doi: str
        DOI to convert

    """
    import urllib

    base_url = "http://dx.doi.org/"
    url = base_url + doi

    req = urllib.request.Request(url)
    req.add_header("Accept", "application/x-bibtex")
    try:
        with urllib.request.urlopen(req) as f:  # noqa: S310
            return f.read().decode()
    except HTTPError as e:
        if e.code == 404:
            print("DOI not found.")
        else:
            print("Service unavailable.")

        return None
    except URLError:
        return None



# flake8: noqa

import os

import numpy as np
import pypolychord
from pypolychord.priors import UniformPrior
from pypolychord.settings import PolyChordSettings

from taurex.util import (
    quantile_corner,
    read_error_into_dict,
    read_error_line,
    read_table,
    recursively_save_dict_contents_to_output,
    weighted_avg_and_std,
)

from .optimizer import Optimizer


class PolyChordOptimizer(Optimizer):
    def __init__(
        self,
        polychord_path=None,
        observed=None,
        model=None,
        num_live_points=1500,
        max_iterations=0,
        maximum_modes=100,
        cluster=True,
        evidence_tolerance=0.5,
        mode_tolerance=-1e90,
        resume=False,
        verbosity=1,
        sigma_fraction=0.1,
    ):
        super().__init__("Multinest", observed, model, sigma_fraction)

        # number of live points
        self.n_live_points = int(num_live_points)
        # maximum no. of iterations (0=inf)
        self.max_iter = int(max_iterations)
        # search for multiple modes
        # parameters on which to cluster, e.g. if nclust_par = 3, it will cluster on the first 3 parameters only.
        # if ncluster_par = -1 it clusters on all parameters
        # maximum number of modes
        self.max_modes = int(maximum_modes)
        self.do_clustering = cluster
        # run in constant efficiency mode
        # set log likelihood tolerance. If change is smaller, multinest will have converged
        self.evidence_tolerance = evidence_tolerance
        self.mode_tolerance = mode_tolerance
        # importance nested sampling

        self.dir_polychord = polychord_path
        self._polychord_output = None
        self.resume = resume
        self.verbose = verbosity

    def compute_fit(self):
        self._polychord_output = None
        data = self._observed.spectrum
        datastd = self._observed.errorBar
        sqrtpi = np.sqrt(2 * np.pi)

        ndim = len(self.fitting_parameters)

        def polychord_loglike(cube):
            # log-likelihood function called by polychord
            fit_params_container = np.array(
                [cube[i] for i in range(len(self.fitting_parameters))]
            )
            chi_t = self.chisq_trans(fit_params_container, data, datastd)

            # print('---------START---------')
            # print('chi_t',chi_t)
            # print('LOG',loglike)
            loglike = -np.sum(np.log(datastd * sqrtpi)) - 0.5 * chi_t
            return loglike, [0.0]

        def polychord_uniform_prior(hypercube):
            # prior distributions called by polychord. Implements a uniform prior
            # converting parameters from normalised grid to uniform prior
            # print(type(cube))
            cube = [0.0] * ndim

            for idx, priors in enumerate(self.fitting_priors):
                # print(idx,self.fitting_parameters[idx])
                cube[idx] = priors.sample(hypercube[idx])
                # print('CUBE idx',cube[idx])
            # print('-----------')
            return cube

        status = None

        datastd_mean = np.mean(datastd)

        settings = PolyChordSettings(ndim, 1)
        settings.nlive = ndim * 25
        settings.num_repeats = ndim * 5
        settings.do_clustering = self.do_clustering
        settings.num_repeats = ndim
        settings.precision_criterion = self.evidence_tolerance
        settings.logzero = -1e70
        settings.read_resume = self.resume
        settings.base_dir = self.dir_polychord
        settings.file_root = "1-"
        self.warning(f"Number of dimensions {ndim}")
        self.warning(f"Fitting parameters {self.fitting_parameters}")

        self.info("Beginning fit......")
        pypolychord.run_polychord(
            polychord_loglike, ndim, 1, settings, polychord_uniform_prior
        )
        self._polychord_output = self.store_polychord_solutions()
        print(self._polychord_output)

    def write_optimizer(self, output):
        opt = super().write_optimizer(output)

        # sampling efficiency (parameter, ...)
        opt.write_scalar("do_clustering", int(self.do_clustering))
        # run in constant efficiency mode
        # set log likelihood tolerance. If change is smaller, multinest will have converged
        opt.write_scalar("evidence_tolerance", self.evidence_tolerance)
        opt.write_scalar("mode_tolerance", self.mode_tolerance)
        # importance nested samplin

        return opt

    def write_fit(self, output):
        fit = super().write_fit(output)

        # if self._polychord_output:
        #     recursively_save_dict_contents_to_output(output,self._polychord_output)

        return fit

    def store_polychord_solutions(self):
        self.warning("Store the polychord results")
        NEST_out = {"solutions": {}}
        data = np.loadtxt(os.path.join(self.dir_polychord, "1-.txt"))

        # self.get_poly_cluster_number(self.dir_polychord)
        NEST_stats = self.get_poly_stats(self.dir_polychord)
        NEST_out["NEST_POLY_stats"] = NEST_stats
        NEST_out["global_logE"] = (
            NEST_out["NEST_POLY_stats"]["global evidence"],
            NEST_out["NEST_POLY_stats"]["global evidence error"],
        )

        modes_array = []
        modes_weights = []
        num_fit_params = len(self.fit_names)
        num_clusters = 1

        if self.do_clustering:
            # if clustering is switched on, checking how many clusters exist
            num_clusters = self.get_poly_cluster_number(self.dir_polychord)
            print("NUM_CLUSTERS: ", num_clusters)
            if num_clusters == 1:
                # Get chains directly from file 1-.txt
                modes_array = [data[:, 2 : num_fit_params + 2]]
                modes_weights = [data[:, 0]]
            else:
                # cycle through clusters
                for midx in range(num_clusters):
                    data = np.loadtxt(
                        os.path.join(self.dir_polychord, f"clusters/1-_{midx + 1}.txt")
                    )
                    modes_array.append(data[:, 2 : num_fit_params + 2])
                    modes_weights.append(data[:, 0])
        else:
            # Get chains directly from file 1-.txt
            modes_array = [data[:, 2 : num_fit_params + 2]]
            modes_weights = [data[:, 0]]

        modes_array = np.asarray(modes_array)
        modes_weights = np.asarray(modes_weights)

        for nmode in range(num_clusters):
            mydict = {
                "type": "nest_poly",
                "local_logE": (
                    NEST_out["NEST_POLY_stats"]["modes"][0]["local log-evidence"],
                    NEST_out["NEST_POLY_stats"]["modes"][0]["local log-evidence error"],
                ),
                "weights": np.asarray(modes_weights[nmode]),
                "tracedata": modes_array[nmode],
                "fit_params": {},
            }

            for idx, param_name in enumerate(self.fit_names):
                trace = modes_array[nmode][:, idx]
                q_16, q_50, q_84 = quantile_corner(
                    trace, [0.16, 0.5, 0.84], weights=np.asarray(modes_weights[nmode])
                )
                mydict["fit_params"][param_name] = {
                    "value": q_50,
                    "sigma_m": q_50 - q_16,
                    "sigma_p": q_84 - q_50,
                    "nest_map": NEST_stats["modes"][nmode]["maximum a posterior"][idx],
                    "nest_mean": NEST_stats["modes"][nmode]["mean"][idx],
                    "nest_sigma": NEST_stats["modes"][nmode]["sigma"][idx],
                    "trace": trace,
                }

            NEST_out["solutions"][f"solution{nmode}"] = mydict

        return NEST_out

    def get_poly_cluster_number(self, dir):
        import glob

        """counts polychord cluster files in 'clusters' folder"""
        cluster_list = glob.glob(os.path.join(dir, "clusters/1-*.txt"))
        c_idx = []
        for file in cluster_list:
            if file[-5].isdigit():
                c_idx.append(int(file[-5]))
        try:
            num = np.max(c_idx)
        except ValueError:
            num = 1
        return num

    def get_poly_stats(self, dir):
        """replicates some of PyMultiNest.Analyzer for PolyChord"""
        stats = {}
        stats["modes"] = {}

        # re-count number of cluster files
        if self.do_clustering:
            num_clusters = self.get_poly_cluster_number(dir)
        else:
            num_clusters = 1

        nmode = 0  # mode index
        # open .stats file and reading global/local evidences
        with open(os.path.join(self.dir_polychord, "1-.stats")) as f:
            lines = f.readlines()
            for idx, line in enumerate(lines):
                if idx == 8:  # skip to global evidence line
                    tmp_line = line.split()
                    gL_mu = tmp_line[2]
                    gL_sig = tmp_line[-1]
                    stats["global evidence"] = gL_mu
                    stats["global evidence error"] = gL_sig
                if idx > 13:  # skip to local evidence
                    tmp_line = line.split()
                    stats["modes"][nmode] = {}
                    stats["modes"][nmode]["local log-evidence"] = tmp_line[3]
                    stats["modes"][nmode]["local log-evidence error"] = tmp_line[5]
                    nmode += 1
                    # stopping reading file when clusters are exhausted
                    if idx == (13 + num_clusters):
                        break

        # opening cluster files (or global file if no clustering) and get MAP, mean, sigma for parameters
        if self.do_clustering:
            for midx in range(num_clusters):
                # cycling through cluster files
                data = np.loadtxt(
                    os.path.join(self.dir_polychord, f"clusters/1-_{midx + 1}.txt")
                )
                # find maximum likelihood index
                mL_idx = np.where(data[:, 1] == np.min(data[:, 1]))
                stats["modes"][midx]["maximum a posterior"] = {}
                stats["modes"][midx]["mean"] = {}
                stats["modes"][midx]["sigma"] = {}
                for idx in range(len(self.fit_names)):
                    # cycle through parameters
                    # maximum likelihood values
                    stats["modes"][midx]["maximum a posterior"][idx] = data[
                        mL_idx, 2 + idx
                    ]
                    # weighted average and sigma
                    mu, sig = weighted_avg_and_std(data[:, 2 + idx], data[:, 0])
                    stats["modes"][midx]["mean"][idx] = mu
                    stats["modes"][midx]["sigma"][idx] = sig

        return stats

    def get_samples(self, solution_idx):
        return self._polychord_output["solutions"][f"solution{solution_idx}"][
            "tracedata"
        ]

    def get_weights(self, solution_idx):
        return self._polychord_output["solutions"][f"solution{solution_idx}"]["weights"]

    def get_solution(self):
        names = self.fit_names
        opt_values = self.fit_values
        opt_map = self.fit_values
        solutions = [
            (k, v)
            for k, v in self._polychord_output["solutions"].items()
            if "solution" in k
        ]

        for k, v in solutions:
            solution_idx = int(k[8:])
            for p_name, p_value in v["fit_params"].items():
                idx = names.index(p_name)
                opt_map[idx] = p_value["nest_map"]
                opt_values[idx] = p_value["value"]

            yield solution_idx, opt_map, opt_values, [
                ("fit_params", v["fit_params"]),
                ("tracedata", v["tracedata"]),
                ("weights", v["weights"]),
            ]

    @classmethod
    def input_keywords(self):
        return [
            "polychord",
            "pypolychord",
        ]

    BIBTEX_ENTRIES = [
        """
        @article{polychord,
            author = {Handley, W. J. and Hobson, M. P. and Lasenby, A. N.},
            title = "{polychord: next-generation nested sampling}",
            journal = {Monthly Notices of the Royal Astronomical Society},
            volume = {453},
            number = {4},
            pages = {4384-4398},
            year = {2015},
            month = {09},
            abstract = "{polychord is a novel nested sampling algorithm tailored for high-dimensional parameter spaces. This paper coincides with the release of polychord v1.6, and provides an extensive account of the algorithm. polychord utilizes slice sampling at each iteration to sample within the hard likelihood constraint of nested sampling. It can identify and evolve separate modes of a posterior semi-independently, and is parallelized using openmpi. It is capable of exploiting a hierarchy of parameter speeds such as those present in cosmomc and camb, and is now in use in the cosmochord and modechord codes. polychord is available for download from http://ccpforge.cse.rl.ac.uk/gf/project/polychord/.}",
            issn = {0035-8711},
            doi = {10.1093/mnras/stv1911},
            url = {https://doi.org/10.1093/mnras/stv1911},
            eprint = {http://oup.prod.sis.lan/mnras/article-pdf/453/4/4384/8034904/stv1911.pdf},
        }
        """,
    ]



"""Chemistry that automatically seperates out active and inactive gases."""
import typing as t

import numpy as np
import numpy.typing as npt

from .chemistry import Chemistry


class AutoChemistry(Chemistry):
    """
    Chemistry class that automatically seperates out
    active and inactive gases

    Has a helper function that should be called at the
    end of initialization. :func:`determine_active_inactive`
    once :func:`gases` has been set.

    You'll only need to implement :func:`gases` and :func:`mixProfile`
    to use this class.

    """

    def __init__(self, name):
        """Initialize auto chemistry.

        Parameters
        ----------

        name: str
            Name of class

        """
        super().__init__(name)

        self._active = []
        self._inactive = []
        self._inactive_mask = None
        self._active_mask = None

    def determine_active_inactive(self) -> None:
        """Determines active and inactive gases."""
        try:
            self._active, self._active_mask = zip(
                *[(m, i) for i, m in enumerate(self.gases) if m in self.availableActive]
            )
        except ValueError:
            self.debug("No active gases detected")
            self._active, self._active_mask = [], None

        try:
            self._inactive, self._inactive_mask = zip(
                *[
                    (m, i)
                    for i, m in enumerate(self.gases)
                    if m not in self.availableActive
                ]
            )
        except ValueError:
            self.debug("No inactive gases detected")
            self._inactive, self._inactive_mask = [], None

        if self._active_mask is not None:
            self._active_mask = np.array(self._active_mask)

        if self._inactive_mask is not None:
            self._inactive_mask = np.array(self._inactive_mask)

    def compute_mu_profile(self, nlayers: t.Optional[int] = None) -> None:
        """Computes molecular weight of atmosphere for each layer

        Parameters
        ----------
        nlayers: int
            Number of layers, deprecated
        """

        if self.mixProfile is not None:
            mix_profile = self.mixProfile
            self.mu_profile = np.sum(
                [
                    self.get_molecular_mass(gas) * mix
                    for gas, mix in zip(self.gases, mix_profile)
                ],
                axis=0,
            )
        else:
            raise ValueError("Mix profile not set/computed")

    @property
    def gases(self) -> t.List[str]:
        """List of gases in atmosphere."""
        raise NotImplementedError

    @property
    def mixProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Mix profile of gases."""
        raise NotImplementedError

    @property
    def activeGases(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Actively absorbing gases."""
        return self._active

    @property
    def inactiveGases(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Non absorbing gases."""
        return self._inactive

    @property
    def activeGasMixProfile(self) -> t.Optional[npt.NDArray[np.float64]]:  # noqa: N802
        """Active gas layer by layer mix profile

        Returns
        -------
        active_mix_profile : :obj:`array`
            Active gas mix profile with shape ``(nactive, nlayer)``


        """

        if self.mixProfile is None:
            raise ValueError("No mix profile computed.")

        if self._active_mask is None:
            return None

        return self.mixProfile[self._active_mask]

    @property
    def inactiveGasMixProfile(  # noqa: N802
        self,
    ) -> t.Optional[npt.NDArray[np.float64]]:
        """Inactive gas layer by layer mix profile.

        Returns
        -------
        inactive_mix_profile : :obj:`array`

        """
        if self.mixProfile is None:
            raise Exception("No mix profile computed.")
        if self._inactive_mask is None:
            return None
        return self.mixProfile[self._inactive_mask]



"""Provides classes related to caching data files needed by taurex 3."""

from .ciaacache import CIACache
from .globalcache import GlobalCache
from .opacitycache import OpacityCache

__all__ = ["OpacityCache", "CIACache", "GlobalCache"]



"""Contains caching class for Molecular cross section files."""

import typing as t

from taurex.core import Singleton
from taurex.log import Logger

from .globalcache import GlobalCache

if t.TYPE_CHECKING:
    from ..opacity import Opacity
else:
    Opacity = object


class OpacityCache(Singleton):
    """Implements a lazy load of opacities.

    A singleton thatloads and caches xsections as they are needed.
    Calling:

    >>> opt = OpacityCache()
    ... opt2 = OpacityCache()

    Reveals that:

    >>> opt == opt2
    True

    Importantly this class will automatically search directories for cross-sections
    set using the :func:`set_opacity_path` method:

    >>> opt.set_opacity_path('path/to/crossections')


    Multiple paths can be set as well

    >>> opt.set_opacity_path(
    ... ['/path/to/crosssections','/another/path/to/crosssections'])

    To get the cross-section object for a particular molecule
    use the square bracket operator:

    >>> opt['H2O']
    <taurex.opacity.pickleopacity.PickleOpacity at 0x107a60be0>

    This returns a :class:`~taurex.opacity.opacity.Opacity` object
    for you to compute H2O cross sections from. When called for the
    first time, a directory search is performed and, if found, the
    appropriate cross-section is loaded. Subsequent calls will immediately
    return the already loaded object:

    >>> h2o_a = opt['H2O']
    >>> h2o_b = opt['H2O']
    >>> h2o_a == h2o_b
    True

    If you have any plugins that include new opacity formats, the cache
    will automatically detect them.



    Lastly you can manually add an opacity directly for a molecule
    into the cache:

    >>> new_h2o = MyNewOpacityFormat()
    >>> new_h2o.molecule
    H2O
    >>> opt.add_opacity(new_h2o)
    >>> opt['H2O']
    <MyNewOpacityFormat at 0x107a60be0>


    Now TauREx3 will use it instead in all calculations!

    """

    def init(self):
        """Initialise the cache."""
        self.opacity_dict: t.Dict[str, Opacity] = {}
        self._opacity_path = None
        self.log = Logger("OpacityCache")
        self._force_active: t.List[str] = []

    def set_opacity_path(self, opacity_path: t.Union[str, t.List[str]]) -> None:
        """Set the path(s) that will be searched for opacities.

        Opacities in this path must be of supported types:

            - HDF5 opacities
            - ``.pickle`` opacities
            - ExoTransmit opacities.

        Parameters
        ----------

        opacity_path : str or :obj:`list` of str, optional
            search path(s) to look for molecular opacities


        Raises
        ------
        NotADirectoryError
            If (any) path does not exist

        """

        import pathlib

        GlobalCache()["xsec_path"] = opacity_path

        paths = [opacity_path] if isinstance(opacity_path, str) else opacity_path
        paths = [pathlib.Path(p) for p in paths]

        for p in paths:
            if not p.exists():
                self.log.error("PATH: %s does not exist!!!", p)
                raise NotADirectoryError
            self.log.debug("Path set to %s", p)

    def enable_radis(self, enable: bool) -> None:
        """Enables/Disables use of RADIS to fill in missing molecules.

        RADIS fills in missing molecules using HITRAN.

        .. warning::
            This is extremely unstable and crashes frequently.
            It is also very slow as it requires
            the computation of the Voigt profile for every temperature.
            We recommend leaving it as False unless necessary.

        Parameters
        ----------
        enable: bool
            Whether to enable RADIS functionality (default = False)

        """

        GlobalCache()["enable_radis"] = enable

    def set_radis_wavenumber(
        self, wn_start: float, wn_end: float, wn_points: int
    ) -> None:
        """Sets the wavenumber grid for RADIS."""
        GlobalCache()["radius_grid"] = wn_start, wn_end, wn_points

        self.clear_cache()

    def set_memory_mode(self, in_memory: bool) -> None:
        """Sets HDF5 opacity load mode.

        For HDF5, whether to stream
        opacities from file (slower, less memory) or load
        them into memory (faster, more memory)

        Parameters
        ----------
        in_memory: bool
            Whether HDF5 files should be streamed (False)
            or loaded into memory (True, default)
        """

        GlobalCache()["xsec_in_memory"] = in_memory
        self.clear_cache()

    def force_active(self, molecules: t.List[str]):
        """Allows some molecules to be forced as active.

        Useful when using other radiative codes to do the calculation

        Parameters
        ----------
        molecules: obj:`list`
            List of molecules

        """
        self._force_active = molecules

    def set_interpolation(self, interpolation_mode: t.Literal["linear", "exp"]):
        """Sets the interpolation mode for cross-sections.

        Can either be ``linear`` for linear interpolation of
        both temeprature and pressure:

        >>> OpacityCache().set_interpolation('linear')

        or ``exp`` for natural exponential interpolation of temperature
        and linear for pressure

        >>> OpacityCache().set_interpolation('exp')


        Parameters
        ----------
        interpolation_mode: str
            Either ``linear`` for bilinear interpolation
            or ``exp`` for exp-linear interpolation.


        """
        GlobalCache()["xsec_interpolation"] = interpolation_mode
        self.clear_cache()

    def __getitem__(self, key: str) -> Opacity:
        """For a molecule return the relevant opacity.

        Returns a :class:`~taurex.opacity.opacity.Opacity` object.

        Parameter
        ---------
        key : str
            molecule name

        Returns
        -------
        :class:`~taurex.opacity.pickleopacity.PickleOpacity`
            Cross-section object desired

        Raises
        ------
        Exception
            If molecule could not be loaded/found


        """
        if key in self.opacity_dict:
            return self.opacity_dict[key]
        else:
            # Try a load of the opacity
            self.load_opacity(molecule_filter=[key])
            # If we have it after a load then good job boys
            if key in self.opacity_dict:
                return self.opacity_dict[key]
            else:
                # Otherwise throw an error
                self.log.error("Opacity for molecule %s could not be loaded", key)
                self.log.error(
                    "It could not be found in the local dictionary %s",
                    list(self.opacity_dict.keys()),
                )
                self.log.error("Or paths %s", GlobalCache()["xsec_path"])
                self.log.error("Try loading it manually/ putting it in a path")
                raise Exception("Opacity could not be loaded")

    def add_opacity(
        self, opacity: Opacity, molecule_filter: t.Optional[t.List[str]] = None
    ):
        """Adds a :class:`~taurex.opacity.opacity.Opacity` object to the cache.

        Parameters
        ----------
        opacity : :class:`~taurex.opacity.opacity.Opacity`
            Opacity object to add to the cache

        molecule_filter : :obj:`list` of str , optional
            If provided, the opacity object will only be included
            if its molecule is in the list. Mostly used by the
            :func:`__getitem__` for filtering

        """
        self.log.info("Reading opacity %s", opacity.moleculeName)
        if opacity.moleculeName in self.opacity_dict:
            self.log.warning(
                "Opacity with name %s already in opactiy dictionary %s skipping",
                opacity.moleculeName,
                self.opacity_dict.keys(),
            )
            return
        if molecule_filter is not None:
            if opacity.moleculeName in molecule_filter:
                self.log.info("Loading opacity %s into model", opacity.moleculeName)
                self.opacity_dict[opacity.moleculeName] = opacity
        else:
            self.log.info("Loading opacity %s into model", opacity.moleculeName)
            self.opacity_dict[opacity.moleculeName] = opacity

    def find_list_of_molecules(self) -> None:
        """Find molecules available to load."""
        from taurex.parameter.classfactory import ClassFactory

        opacity_klasses = ClassFactory().opacityKlasses

        molecules = []

        for c in opacity_klasses:
            molecules.extend([x[0] for x in c.discover()])

        forced = self._force_active or []
        return set(molecules + forced + list(self.opacity_dict.keys()))

    def load_opacity_from_path(
        self, path: str, molecule_filter: t.Optional[t.List[str]] = None
    ):
        """Searches path for molecular cross-section files.

        Creates and loads opacities into the cache
        ``.pickle`` will be loaded as
        :class:`~taurex.opacity.pickleopacity.PickleOpacity`


        Parameters
        ----------
        path : str
            Path to search for molecular cross-section files

        molecule_filter : :obj:`list` of str , optional
            If provided, the opacity will only be loaded
            if its molecule is in this list. Mostly used by the
            :func:`__getitem__` for filtering

        """
        from taurex.parameter.classfactory import ClassFactory

        cf = ClassFactory()

        opacity_klass_list = sorted(cf.opacityKlasses, key=lambda x: x.priority())

        for c in opacity_klass_list:
            for mol, args in c.discover():
                self.log.debug("Klass: %s %s", mol, args)
                op = None
                if mol in molecule_filter and mol not in self.opacity_dict:
                    if not isinstance(
                        args,
                        (
                            list,
                            tuple,
                        ),
                    ):
                        args = [args]
                    op = c(*args)

                if op is not None and op.moleculeName not in self.opacity_dict:
                    self.add_opacity(op, molecule_filter=molecule_filter)
                op = None  # Ensure garbage collection when run once

    def load_opacity(
        self,
        opacities: t.Optional[t.Union[Opacity, t.List[Opacity]]] = None,
        opacity_path: t.Union[str, t.List[str]] = None,
        molecule_filter: t.Optional[t.List[str]] = None,
    ):
        """Main function to use when loading molecular opacities.
        Handles both cross sections and paths. Handles lists of either so lists of
        :class:`~taurex.opacity.opacity.Opacity` objects or lists of paths can be used
        to load multiple files/objects


        Parameters
        ----------
        opacities : :class:`~taurex.opacity.opacity.Opacity` or
        :obj:`list` of :class:`~taurex.opacity.opacity.Opacity` , optional
            Object(s) to include in cache

        opacity_path : str or :obj:`list` of str, optional
            search path(s) to look for molecular opacities

        molecule_filter : :obj:`list` of str , optional
            If provided, the opacity will only be loaded
            if its molecule is in this list. Mostly used by the
            :func:`__getitem__` for filtering

        """
        from taurex.opacity import Opacity

        if opacity_path is None:
            opacity_path = GlobalCache()["xsec_path"]

        if opacities is not None:
            if isinstance(opacities, (list,)):
                self.log.debug("Opacity passed is list")
                for opacity in opacities:
                    self.add_opacity(opacity, molecule_filter=molecule_filter)
            elif isinstance(opacities, Opacity):
                self.add_opacity(opacities, molecule_filter=molecule_filter)
            else:
                self.log.error(
                    "Unknown type %s passed into opacities, should be a list, single \
                     opacity or None if reading a path",
                    type(opacities),
                )
                raise Exception("Unknown type passed into opacities")
        else:
            self.load_opacity_from_path(opacity_path, molecule_filter=molecule_filter)

    def clear_cache(self) -> None:
        """Clears all currently loaded cross-sections."""
        self.opacity_dict = {}



"""Modules for handling pressure."""
from .arraypressure import ArrayPressureProfile
from .filepressure import FilePressureProfile
from .pressureprofile import LogPressureProfile, PressureProfile, SimplePressureProfile

__all__ = [
    "SimplePressureProfile",
    "PressureProfile",
    "LogPressureProfile",
    "ArrayPressureProfile",
    "FilePressureProfile",
]



"""These modules handle the loading of collisionaly induced absorption data files."""


from .cia import CIA
from .hitrancia import HitranCIA
from .picklecia import PickleCIA  # noqa: S403

__all__ = ["CIA", "HitranCIA", "PickleCIA"]



"""Core classes for Taurex."""
from taurex.data.citation import Citable, to_bibtex, unique_citations_only
from taurex.data.fittable import (
    DerivedType,
    Fittable,
    FittingType,
    derivedparam,
    fitparam,
)
from taurex.output.output import Output

"""Just contains a singleton class. Pretty useful"""


class Singleton:
    """
    A singleton for your usage. When inheriting do not implement __init__ instead
    override :func:`init`


    """

    def __new__(cls, *args, **kwds):
        it = cls.__dict__.get("__it__")
        if it is not None:
            return it
        cls.__it__ = it = object.__new__(cls)
        it.init(*args, **kwds)
        return it

    def init(self, *args, **kwds):
        """Override to act as an init"""
        pass


__all__ = [
    "fitparam",
    "derivedparam",
    "Fittable",
    "Citable",
    "unique_citations_only",
    "to_bibtex",
    "Output",
    "Singleton",
    "FittingType",
    "DerivedType",
]



"""Constant gas profile."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.output import OutputGroup
from taurex.util import molecule_texlabel

from .gas import Gas


class ConstantGas(Gas):
    """Constant gas profile.

    Molecular abundace is constant at each layer of the
    atmosphere

    """

    def __init__(
        self,
        molecule_name: t.Optional[str] = "H2O",
        mix_ratio: t.Optional[float] = 1e-5,
    ) -> None:
        """Initialize constant gas profile.

        Parameters
        -----------
        molecule_name : str
            Name of molecule

        mix_ratio : float
            Mixing ratio of the molecule


        """
        super().__init__("ConstantGas", molecule_name)
        self._mix_ratio = mix_ratio
        self.add_active_gas_param()

    @property
    def mixProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """

        Mixing profile

        Returns
        -------
        mix: :obj:`array`
            Mix ratio for molecule at each layer

        """

        return self._mix_array

    def initialize_profile(
        self,
        nlayers: t.Optional[int] = None,
        temperature_profile: t.Optional[npt.NDArray[np.float64]] = None,
        pressure_profile: t.Optional[npt.NDArray[np.float64]] = None,
        altitude_profile: t.Optional[npt.NDArray[np.float64]] = None,
    ) -> None:
        """Initialize the mixing profile.

        Parameters
        -----------
        nlayers: int
            Number of layers in atmosphere
        temperature_profile: :obj:`array`
            Temperature profile of atmosphere
        pressure_profile: :obj:`array`
            Pressure profile of atmosphere
        altitude_profile: :obj:`array`
            Altitude profile of atmosphere, deprecated

        """
        self._mix_array = np.full(nlayers, self._mix_ratio)

    def add_active_gas_param(self) -> None:
        """Add the mixing ratio as a fitting parameter.

        Fitting parameter identifier is the molecule name.
        Generates a fitting parameter on the fly by building
        getter and setter functions and passing them to
        :func:`taurex.fitting.fittable.Fittable.add_fittable_param`

        """

        mol_name = self.molecule
        param_name = self.molecule
        param_tex = molecule_texlabel(mol_name)

        def read_mol(self):
            return self._mix_ratio

        def write_mol(self, value):
            self._mix_ratio = value

        read_mol.__doc__ = f"{mol_name} constant mix ratio (VMR)"

        fget = read_mol
        fset = write_mol

        bounds = [1.0e-12, 0.1]

        default_fit = False
        self.add_fittable_param(
            param_name, param_tex, fget, fset, "log", default_fit, bounds
        )

    def write(self, output: OutputGroup):
        """Write constant gas profile to output."""
        gas_entry = super().write(output)
        gas_entry.write_scalar("mix_ratio", self._mix_ratio)

        return gas_entry

    @classmethod
    def input_keywords(cls) -> t.Tuple[str]:
        return ("constant",)



"""The main taurex program"""


def parse_keywords(keywords):
    import tabulate

    from taurex.parameter.classfactory import ClassFactory

    cf = ClassFactory()
    print("\n")
    if keywords in ("contribs",):
        print("")
        print("-----------------------------------------------")
        print("-------------Available Contributions-----------")
        print("-----------------------------------------------")
        print("")
        table = [
            (f"[[{c.__name__}]]", c.__module__.split(".")[0].split("_")[-1])
            for c in cf.contributionKlasses
            if hasattr(c, "input_keywords")
        ]
        print(
            tabulate.tabulate(
                table, headers=["Header", "Source"], tablefmt="fancy_grid"
            )
        )

    elif keywords in ("chemistry",):
        print("")
        print("-----------------------------------------------")
        print("-------------Available [Chemistry]-------------")
        print("-----------------------------------------------")
        print("")
        table = [
            (
                " / ".join(c.input_keywords()),
                f"{c.__name__}",
                c.__module__.split(".")[0].split("_")[-1],
            )
            for c in cf.chemistryKlasses
            if hasattr(c, "input_keywords")
        ]
        print(
            tabulate.tabulate(
                table,
                headers=["chemistry_type", "Class", "Source"],
                tablefmt="fancy_grid",
            )
        )
        print("\n")
    elif keywords in ("temperature",):
        print("")
        print("-----------------------------------------------")
        print("-------------Available [Temperature]-----------")
        print("-----------------------------------------------")
        print("")
        table = [
            (
                " / ".join(c.input_keywords()),
                f"{c.__name__}",
                c.__module__.split(".")[0].split("_")[-1],
            )
            for c in cf.temperatureKlasses
            if hasattr(c, "input_keywords")
        ]
        print(
            tabulate.tabulate(
                table,
                headers=["profile_type", "Class", "Source"],
                tablefmt="fancy_grid",
            )
        )
        print("\n")
    elif keywords in ("gas",):
        print("")
        print("-----------------------------------------------")
        print("-------------Available Gas Profiles------------")
        print("-----------------------------------------------")
        print("")
        table = [
            (
                " / ".join(c.input_keywords()),
                f"{c.__name__}",
                c.__module__.split(".")[0].split("_")[-1],
            )
            for c in cf.gasKlasses
            if hasattr(c, "input_keywords")
        ]

        print(
            tabulate.tabulate(
                table, headers=["gas_type", "Class", "Source"], tablefmt="fancy_grid"
            )
        )
        print("\n")
    elif keywords in ("optimizer",):
        print("")
        print("-----------------------------------------------")
        print("-------------Available Optimizers--------------")
        print("-----------------------------------------------")
        print("")
        table = [
            (
                " / ".join(c.input_keywords()),
                f"{c.__name__}",
                c.__module__.split(".")[0].split("_")[-1],
            )
            for c in cf.optimizerKlasses
            if hasattr(c, "input_keywords")
        ]
        print(
            tabulate.tabulate(
                table, headers=["optimizer", "Class", "Source"], tablefmt="fancy_grid"
            )
        )
        print("\n")
    elif keywords in ("prior",):
        print("")
        print("-----------------------------------------------")
        print("-------------Available Priors------------------")
        print("-----------------------------------------------")
        print("")
        table = [
            (f"{c.__name__}", c.__module__.split(".")[0].split("_")[-1])
            for c in cf.priorKlasses
            if hasattr(c, "input_keywords")
        ]
        print(
            tabulate.tabulate(
                table, headers=["prior", "Class", "Source"], tablefmt="fancy_grid"
            )
        )
        print("\n")
    elif keywords in ("model",):
        print("")
        print("-----------------------------------------------")
        print("-------------Available Forward [Model]s--------")
        print("-----------------------------------------------")
        print("")
        table = [
            (
                " / ".join(c.input_keywords()),
                f"{c.__name__}",
                c.__module__.split(".")[0].split("_")[-1],
            )
            for c in cf.modelKlasses
            if hasattr(c, "input_keywords")
        ]
        print(
            tabulate.tabulate(
                table, headers=["model_type", "Class", "Source"], tablefmt="fancy_grid"
            )
        )
        print("\n")
    elif keywords in ("pressure",):
        print("")
        print("-----------------------------------------------")
        print("-------------Available [Pressure]s-------------")
        print("-----------------------------------------------")
        print("")
        table = [
            (
                " / ".join(c.input_keywords()),
                f"{c.__name__}",
                c.__module__.split(".")[0].split("_")[-1],
            )
            for c in cf.pressureKlasses
        ]
        print(
            tabulate.tabulate(
                table,
                headers=["profile_type", "Class", "Source"],
                tablefmt="fancy_grid",
            )
        )
        print("\n")


def show_parameters(model):
    import tabulate

    print("")
    print("-----------------------------------------------")
    print("------Available Retrieval Parameters-----------")
    print("-----------------------------------------------")
    print("")

    keywords = [k for k, v in model.fittingParameters.items()]

    short_desc = []
    for v in model.fittingParameters.values():
        doc = v[2].__doc__
        if doc is None or doc == "None":
            short_desc.append("")
        else:
            split = doc.split("\n")
            for spl in split:
                if len(spl) > 0:
                    s = spl
                    break

            short_desc.append(s)

    output = tabulate.tabulate(
        zip(keywords, short_desc),
        headers=["Param Name", "Short Desc"],
        tablefmt="fancy_grid",
    )
    print(output)
    print("\n\n")

    import tabulate

    print("")
    print("-----------------------------------------------")
    print("------Available Computable Parameters----------")
    print("-----------------------------------------------")
    print("")

    keywords = [k for k, v in model.derivedParameters.items()]

    short_desc = []
    for v in model.derivedParameters.values():
        doc = v[2].__doc__
        if doc is None or doc == "None":
            short_desc.append("")
        else:
            split = doc.split("\n")
            for spl in split:
                if len(spl) > 0:
                    s = spl
                    break

            short_desc.append(s)

    output = tabulate.tabulate(
        zip(keywords, short_desc),
        headers=["Param Name", "Short Desc"],
        tablefmt="fancy_grid",
    )
    print(output)
    print("\n\n")


def show_plugins():
    import logging

    from taurex.log import setLogLevel
    from taurex.parameter.classfactory import ClassFactory

    setLogLevel(logging.ERROR)

    successful_plugins, failed_plugins = ClassFactory().discover_plugins()

    print("\nSuccessfully loaded plugins")
    print("---------------------------")
    for k in successful_plugins.keys():
        print(k)

    print("\n\nFailed plugins")
    print("---------------------------")
    for k, v in failed_plugins.items():
        print(k)
        print(f"Reason: {v}")

    print("\n")


def output_citations(model, instrument, optimizer):  # noqa: C901
    from taurex.cache import OpacityCache
    from taurex.mpi import barrier, get_rank

    barrier()
    bib_tex = None
    citation_string = None
    if get_rank() == 0:
        print("\n\n----------------------------------------------------------")
        print("----------------------Bibiliography-----------------------")
        print("----------------------------------------------------------")

        print("If you use any of the results from this run please cite")
        print("the following publications:")

        citation_string = ""
        all_citations = []
        print("\n")
        print("TauREx-Related")
        print("--------------\n")
        from taurex._citation import __citations__, taurex_citation

        citation_string += __citations__
        all_citations.extend(taurex_citation.citations())
        print(__citations__)

        print("Forward model")
        print("-------------\n")

        cite = model.nice_citation()
        all_citations.extend(model.citations())
        citation_string += cite
        print(cite)

        if optimizer is not None:
            cite = optimizer.nice_citation()
            all_citations.extend(optimizer.citations())
            if len(cite) > 0:
                citation_string += cite
                print("Optimizer")
                print("---------\n")
                print(cite)

        if instrument is not None:
            cite = instrument.nice_citation()
            all_citations.extend(instrument.citations())
            if len(cite) > 0:
                citation_string += cite
                print("Instrument")
                print("---------\n")
                print(cite)

        oc = OpacityCache()

        active_gases = model.chemistry.activeGases

        # Do cross sections

        missing_citations = []

        for g in active_gases:
            try:
                xsec = oc.opacity_dict[g]
                if len(xsec.opacityCitation()) == 0:
                    missing_citations.append(g)
            except KeyError:
                continue

        if missing_citations:
            print("Missing Opacity Citations")
            print("------------------------\n")
            print("These opacities do not have citation information, please")
            print("determine their appropriate publications and cite them \n")
            print(missing_citations)
            print("\n")
        from taurex.core import to_bibtex

        bib_tex = to_bibtex([a for a in all_citations if not isinstance(a, str)])

    barrier()

    return bib_tex, citation_string


def only_bibtex(filename, pp):
    model = pp.generate_appropriate_model()
    instrument = pp.generate_instrument()[0]
    optimizer = pp.generate_optimizer()

    bib_tex, citation_string = output_citations(model, instrument, optimizer)
    if bib_tex:
        with open(filename, "w") as f:
            f.write(bib_tex)


def main():  # noqa: C901
    import argparse
    import datetime
    import logging

    import numpy as np

    from taurex.log import setLogLevel
    from taurex.log.logger import TauRexHandler, root_logger
    from taurex.mpi import get_rank
    from taurex.output.hdf5 import HDF5Output
    from taurex.parameter import ParameterParser
    from taurex.util.output import store_contributions

    from . import __version__ as version
    from .taurexdefs import OutputSize

    rh = TauRexHandler()
    formatter = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
    rh.setFormatter(formatter)
    rh.setLevel(logging.DEBUG)
    root_logger.handlers = []
    root_logger.addHandler(rh)
    root_logger.setLevel(logging.INFO)

    parser = argparse.ArgumentParser(description=f"TauREx {version}")

    parser.add_argument(
        "-i", "--input", dest="input_file", type=str, help="Input par file to pass"
    )

    parser.add_argument(
        "-R",
        "--retrieval",
        dest="retrieval",
        default=False,
        help="When set, runs retrieval",
        action="store_true",
    )

    parser.add_argument(
        "-p",
        "--plot",
        dest="plot",
        default=False,
        help="Whether to plot after the run",
        action="store_true",
    )

    parser.add_argument(
        "-g",
        "--debug-log",
        dest="debug",
        default=False,
        help="Debug log output",
        action="store_true",
    )

    parser.add_argument(
        "-c",
        "--show-contrib",
        dest="contrib",
        default=False,
        help="Show basic contributions",
        action="store_true",
    )

    parser.add_argument(
        "-C",
        "--full-contrib",
        dest="full_contrib",
        default=False,
        help="Show ALL contributions",
        action="store_true",
    )

    parser.add_argument(
        "--light",
        dest="light",
        default=False,
        help="Light outputs",
        action="store_true",
    )

    parser.add_argument(
        "--lighter",
        dest="lighter",
        default=False,
        help="Even Lighter outputs",
        action="store_true",
    )

    parser.add_argument("-o", "--output_file", dest="output_file", type=str)

    parser.add_argument("-S", "--save-spectrum", dest="save_spectrum", type=str)

    parser.add_argument(
        "-v",
        "--version",
        dest="version",
        default=False,
        help="Display version",
        action="store_true",
    )

    parser.add_argument(
        "--plugins",
        dest="plugins",
        default=False,
        help="Display plugins",
        action="store_true",
    )

    parser.add_argument(
        "--fitparams",
        dest="fitparams",
        default=False,
        help="Display available fitting params",
        action="store_true",
    )

    parser.add_argument(
        "--bibtex", dest="bibtex", type=str, help="Output bibliography .bib to filepath"
    )

    parser.add_argument(
        "--only-bib",
        dest="no_run",
        help="Do not run anything, only store bibtex (must have --bibtex)",
        default=False,
        action="store_true",
    )

    parser.add_argument("--keywords", dest="keywords", type=str)

    args = parser.parse_args()

    output_size = OutputSize.heavy

    if args.version:
        print(version)
        return

    if args.plugins:
        show_plugins()
        return

    if args.keywords:
        parse_keywords(args.keywords)
        return

    if args.input_file is None:
        print("Fatal: No input file specified.")
        return

    if args.light:
        output_size = OutputSize.light

    if args.lighter:
        output_size = OutputSize.lighter

    if args.debug:
        setLogLevel(logging.DEBUG)

    root_logger.info("TauREx %s", version)

    root_logger.info("TauREx PROGRAM START AT %s", datetime.datetime.now())

    # Parse the input file
    pp = ParameterParser()
    pp.read(args.input_file)

    # Setup global parameters
    pp.setup_globals()

    if args.no_run and args.bibtex:
        return only_bibtex(args.bibtex, pp)

    # Get the spectrum
    observation = pp.generate_observation()

    binning = pp.generate_binning()

    # Generate a model from the input
    model = pp.generate_appropriate_model(obs=observation)

    # build the model
    model.build()

    if args.fitparams:
        show_parameters(model)
        return

    wngrid = None

    if binning == "observed" and observation is None:
        logging.critical("Binning selected from Observation yet None provided")
        quit()

    if binning is None:
        if observation is None or observation == "self":
            binning = model.defaultBinner()
            wngrid = model.nativeWavenumberGrid
        else:
            binning = observation.create_binner()
            wngrid = observation.wavenumberGrid
    else:
        if binning == "native":
            binning = model.defaultBinner()
            wngrid = model.nativeWavenumberGrid
        elif binning == "observed":
            binning = observation.create_binner()
            wngrid = observation.wavenumberGrid
        else:
            binning, wngrid = binning

    instrument = pp.generate_instrument(binner=binning)

    num_obs = 1
    if instrument is not None:
        instrument, num_obs = instrument

    if observation == "self" and instrument is None:
        logging.getLogger("taurex").critical(
            "Instrument nust be specified when using self option"
        )
        raise ValueError("No instruemnt specified for self option")

    inst_result = None
    if instrument is not None:
        inst_result = instrument.model_noise(
            model, model_res=model.model(), num_observations=num_obs
        )

    # Observation on self
    if observation == "self":
        from taurex.data.spectrum import ArraySpectrum
        from taurex.util import wnwidth_to_wlwidth

        inst_wngrid, inst_spectrum, inst_noise, inst_width = inst_result

        inst_wlgrid = 10000 / inst_wngrid

        inst_wlwidth = wnwidth_to_wlwidth(inst_wngrid, inst_width)
        observation = ArraySpectrum(
            np.vstack([inst_wlgrid, inst_spectrum, inst_noise, inst_wlwidth]).T
        )
        binning = observation.create_binner()

    # Handle outputs
    if args.output_file:
        # Output taurex data
        with HDF5Output(args.output_file) as o:
            model.write(o)

    optimizer = None
    solution = None

    model.model()
    if args.retrieval is True:
        import time

        if observation is None:
            logging.critical("No spectrum is defined!!")
            quit()

        optimizer = pp.generate_optimizer()
        optimizer.set_model(model)
        optimizer.set_observed(observation)
        pp.setup_optimizer(optimizer)

        start_time = time.time()
        solution = optimizer.fit(output_size=output_size)

        end_time = time.time()

        root_logger.info("Total Retrieval finish in %s seconds", end_time - start_time)

        for _, optimized, _, _ in optimizer.get_solution():
            optimizer.update_model(optimized)
            break

    result = model.model()

    if args.save_spectrum is not None:
        # with open(args.save_spectrum, 'w') as f:
        from taurex.util import compute_bin_edges, wnwidth_to_wlwidth

        save_wnwidth = compute_bin_edges(wngrid)[1]
        save_wl = 10000 / wngrid
        save_wlwidth = wnwidth_to_wlwidth(wngrid, save_wnwidth)
        save_model = binning.bin_model(result)[1]
        save_error = np.zeros_like(save_wl)
        if inst_result is not None:
            inst_wngrid, inst_spectrum, inst_noise, inst_width = inst_result

            save_model = inst_spectrum
            save_wl = 10000 / inst_wngrid

            save_wlwidth = wnwidth_to_wlwidth(inst_wngrid, inst_width)

            save_error = inst_noise

        np.savetxt(
            args.save_spectrum,
            np.vstack((save_wl, save_model, save_error, save_wlwidth)).T,
        )

    if args.output_file:
        # Output taurex data
        with HDF5Output(args.output_file, append=True) as o:
            out = o.create_group("Output")
            if observation is not None:
                obs = o.create_group("Observed")
                observation.write(obs)

            profiles = model.generate_profiles()
            spectrum = binning.generate_spectrum_output(result, output_size=output_size)

            if inst_result is not None:
                spectrum["instrument_wngrid"] = inst_result[0]
                spectrum["instrument_wnwidth"] = inst_result[-1]
                spectrum["instrument_wlgrid"] = 10000 / inst_result[0]
                spectrum["instrument_spectrum"] = inst_result[1]
                spectrum["instrument_noise"] = inst_result[2]

            try:
                spectrum["Contributions"] = store_contributions(
                    binning, model, output_size=output_size - 3
                )
            except Exception:  # noqa: S110
                pass

            if solution is not None:
                out.store_dictionary(solution, group_name="Solutions")
                priors = {}
                priors["Profiles"] = profiles
                priors["Spectra"] = spectrum
                out.store_dictionary(priors, group_name="Priors")
            else:
                out.store_dictionary(profiles, group_name="Profiles")
                out.store_dictionary(spectrum, group_name="Spectra")

            if optimizer:
                optimizer.write(o)

    bib_tex, citation_string = output_citations(model, instrument, optimizer)

    if args.output_file and bib_tex and citation_string:
        with HDF5Output(args.output_file, append=True) as o:
            bib = o.create_group("Bibliography")
            bib.write_string("short_form", citation_string)
            bib.write_string("bibtex", bib_tex)

    if args.bibtex and bib_tex and citation_string:
        with open(args.bibtex, "w") as f:
            f.write(bib_tex)

    root_logger.info("TauREx PROGRAM END AT %s s", datetime.datetime.now())

    if args.plot:
        wlgrid = 10000 / wngrid
        if get_rank() == 0:
            import matplotlib.pyplot as plt

            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)

            ax.set_xlabel(r"Wavelength ($\mu$m)")
            ax.set_ylabel(r"$(R_p/R_s)^2$")
            is_lightcurve = False
            try:
                from taurex.model.lightcurve.lightcurve import LightCurveModel

                is_lightcurve = isinstance(model, LightCurveModel)
                ax.set_xscale("linear")
            except ImportError:
                pass

            if observation is not None:
                if is_lightcurve:
                    ax.plot(observation.spectrum.flatten(), label="observation")
                else:
                    ax.errorbar(
                        observation.wavelengthGrid,
                        observation.spectrum,
                        observation.errorBar,
                        fmt=".",
                        label="observation",
                    )

            if is_lightcurve:
                ax.plot(result[1], label="forward model")
            else:
                if inst_result is not None:
                    from taurex.util import wnwidth_to_wlwidth

                    inst_wngrid, inst_spectrum, inst_noise, inst_width = inst_result

                    inst_wlgrid = 10000 / inst_wngrid

                    inst_wlwidth = wnwidth_to_wlwidth(inst_wngrid, inst_width)

                    ax.errorbar(
                        inst_wlgrid,
                        inst_spectrum,
                        inst_noise,
                        inst_wlwidth / 2,
                        ".",
                        label="Instrument",
                    )

                else:
                    ax.plot(wlgrid, binning.bin_model(result)[1], label="forward model")

                ax.set_xscale("log")

            if args.contrib:
                native_grid, contrib_result = model.model_contrib(wngrid=wngrid)

                for contrib_name, contrib in contrib_result.items():
                    flux, tau, extras = contrib

                    binned = binning.bindown(native_grid, flux)
                    if is_lightcurve:
                        ax.plot(binned[1], label=contrib_name)
                    else:
                        ax.plot(wlgrid, binned[1], label=contrib_name)

            if args.full_contrib:
                native_grid, contrib_result = model.model_full_contrib(wngrid=wngrid)

                for contrib_name, contrib in contrib_result.items():
                    for name, flux, _, _ in contrib:
                        label = f"{contrib_name} - {name}"

                        binned = binning.bindown(native_grid, flux)
                        if is_lightcurve:
                            ax.plot(binned[1], label=label)
                        else:
                            ax.plot(wlgrid, binned[1], label=label)

            plt.legend()
            plt.show()
        else:
            logging.getLogger("taurex").warning(
                "Number of processes > 1 so not plotting"
            )


if __name__ == "__main__":
    main()
